#include "bmp.h"
#include <stdint.h>
#include <iostream>     // std::cout
#include <fstream>      // std::ifstream

imageBMP::imageBMP(const char *name /** INPUT FILE NAME */ ){

    width_=height_=size_=0;

   /// WE CHECK THE EXTENSION OF THE FILE
   //printf("strlen(%s)=%d\n",name,strlen(name));
    int N=strlen(name);
    if(N<4) return;
    if( (name[N-3]=='p' || name[N-3]=='P') && (name[N-2]=='n' || name[N-2]=='N') && (name[N-1]=='g' || name[N-1]=='G') ){
      /// PNG FORMAT FILE
      vector<unsigned char> image;
      unsigned int width0,height0;
      unsigned int error = lodepng::decode(image, width0, height0,name);
      if(error){
        printf("Problems reading image %s\n",name);
        return;
      }
      width_=width0;
      height_=height0;
      size_=width0*height0;
      RGB_.resize(size_);

      if(image.size()!=(unsigned int) 4*size_){
        printf("Problems with dimension of image %s\n",name);
        printf("size_=%d image.size()/4=%d\n",size_,(int) image.size()/4);
        return;
      }

      unsigned int cont=0;
      for(int y=height_-1;y>=0;y--){
        for(int x=0;x<width_;x++){
          unsigned int k=y*width_+x;
          RGB_[k].B_=image[cont++];
          RGB_[k].G_=image[cont++];
          RGB_[k].R_=image[cont++];
          cont++;
        }
      }
      return;
   }

   /// BMP FORMAT FILE

   HEADER header;
   INFOHEADER infoheader;
   FILE *fptr;
   long size,m,cont;
   unsigned char *image;


   /* Open file */
   if ((fptr = fopen(name,"rb")) == NULL) {
      fprintf(stderr,"Unable to open BMP file \"%s\"\n",name);
      return;
   }

   /* Read and check the header */
   if (fread(&header.type,sizeof(unsigned short),1,fptr) != 1) {
       fprintf(stderr,"Image read failed\n");
       return;
   }

   /* Read and check the header */
   //fprintf(stderr,"ID is: %d, should be %d\n",header.type,'M'*256+'B');
   if (header.type != (int) 'M'*256+'B') {
     printf("Be careful, the file is not a bmp image file (magic number not correct)\n");
     //fprintf(stderr,"Be careful, the file is not a bmp image file (magic number not correct)\n");
     //fprintf(stderr,"  File : %s \n",name);
     //scanf("%d",&cont);
   }

   if (fread(&header.size,sizeof(unsigned int),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }

   //fprintf(stderr,"File size is %d bytes\n",header.size);
   if (fread(&header.reserved1,sizeof(unsigned short),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }

   if (fread(&header.reserved2,sizeof(unsigned short),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }
   //fprintf(stderr,"header.reserved1 = %d \n",header.reserved1);
   if (fread(&header.offset,sizeof(unsigned int),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
    }
    //fprintf(stderr,"header.reserved2 = %d \n",header.reserved2);
    //fprintf(stderr,"Offset to image data is %d bytes\n",header.offset);

   /* Read and check the information header */
   if (fread(&infoheader,sizeof(INFOHEADER),1,fptr) != 1) {
      fprintf(stderr,"Failed to read BMP info header\n");
      return;
   }
//   fprintf(stderr,"Header size = %d \n",infoheader.size);
//   fprintf(stderr,"Image size = %d x %d\n",infoheader.width,infoheader.height);
//   fprintf(stderr,"Number of colour planes is %d\n",infoheader.planes);
//   fprintf(stderr,"Bits per pixel is %d\n",infoheader.bits);
//   fprintf(stderr,"Compression type is %d\n",infoheader.compression);
//   fprintf(stderr,"Infoheader.imagesize is %d\n",infoheader.imagesize);
//   fprintf(stderr,"Infoheader.xresolution=%d infoheader.yresolution=%d \n",
//           infoheader.xresolution,infoheader.yresolution);
//   fprintf(stderr,"Number of colours is %d\n",infoheader.ncolours);
//   fprintf(stderr,"Number of required colours is %d\n",
//      infoheader.importantcolours);

   /* WE CHECK THE 24 bits */
   if(infoheader.bits!=24){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }
   /* WE CHECK COMPRESSION */
   if(infoheader.compression!=0){
	   fprintf(stderr,"Sorry, function only implemented image without compression \n");
       return;
   }
   /* WE CHECK THE 24 bits */
   if(infoheader.ncolours!=0){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }
   /* WE CHECK THE 24 bits */
   if(infoheader.importantcolours!=0){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }


   /* WE FIT DIMENSIONS AND ALLOCATE MEMORY FOR THE IMAGE */
   width_=infoheader.width; height_=infoheader.height;
   size=width_*height_;
   RGB_.resize(size);

  /* WE STORE THE IMAGE INFORMATION IN image */
   image=(unsigned char*)malloc(3*size*sizeof(unsigned char));

   /* We read the image */
   if (fread(image,size*3,1,fptr) != 1) {
     fprintf(stderr,"Image read failed\n");
     free(image);
     return;
   }

   cont=0;
   for(m=0;m<size;m++){
       RGB_[m].R_=image[cont];
       RGB_[m].G_=image[cont+1];
       RGB_[m].B_=image[cont+2]; cont+=3;
   }
   fclose(fptr);

}


int imageBMP::write(const char *name){

    /// WE CHECK THE EXTENSION OF THE FILE
    //printf("strlen(%s)=%d\n",name,strlen(name));
    int N=strlen(name);
    if(N<4) return -1;
    if( (name[N-3]=='p' || name[N-3]=='P') && (name[N-2]=='n' || name[N-2]=='N') && (name[N-1]=='g' || name[N-1]=='G') ){
      /// PNG FORMAT FILE
      vector<unsigned char> image((unsigned int) width_ * height_ * 4);
      unsigned int cont=0;
      for(int y=height_-1;y>=0;y--){
        for(int x=0;x<width_;x++){
          unsigned int k=y*width_+x;
          image[cont++]=(unsigned int) RGB_[k].B_<0?0:RGB_[k].B_>255?255:(round(RGB_[k].B_));
          image[cont++]=(unsigned int) RGB_[k].G_<0?0:RGB_[k].G_>255?255:(round(RGB_[k].G_));
          image[cont++]=(unsigned int) RGB_[k].R_<0?0:RGB_[k].R_>255?255:(round(RGB_[k].R_));
          image[cont++]=255;
        }
      }
      unsigned error = lodepng::encode(name, image, (unsigned int) width_, (unsigned int) height_);

      if(error){
        printf("Problems writing image %s\n",name);
        return-1;
      }

      return 0;

    }



   /// BMP FORMAT FILE
   int width=4*(width_/4);
   int height=4*(height_/4);
   if(width==0 || height==0) return -1;
   if(width!=width_ || height!=height_){
     imageBMP I2(width,height);
     for(int x=0;x<width;x++){
       for(int y=0;y<height;y++){
         I2[y*width+x]=RGB_[y*width_+x];
       }
     }
     return I2.write(name);
   }

   HEADER header;
   INFOHEADER infoheader;
   FILE *fptr;
   long cont,size,m;
   unsigned char *image;

   /* Open file */
   if ((fptr = fopen(name,"wb")) == NULL) {
      fprintf(stderr,"Unable to open BMP file \"%s\"\n",name);
      return(-1);
   }

   /* Header Definition */
   header.type = (unsigned short int) 'M'*256+'B';
   header.size = (unsigned int) 3*width_*height_+54;
   header.reserved1 = 0;
   header.reserved2 = 0;
   header.offset = (unsigned int) 54;


  /* We write the header */
   if (fwrite(&header.type,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.size,sizeof(unsigned int),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.reserved1,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.reserved2,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.offset,sizeof(unsigned int),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }


   /* INFOHEADER DEFINITION */
   infoheader.size = 40;
   infoheader.width = width_;
   infoheader.height = height_;
   infoheader.planes = (unsigned short int) 1;
   infoheader.bits = (unsigned short int) 24;
   infoheader.compression = (unsigned int) 0;
   infoheader.imagesize=3*width_*height_;
   infoheader.xresolution=2834;
   infoheader.yresolution=2834;
   infoheader.ncolours = (unsigned int) 0 ;
   infoheader.importantcolours = (unsigned int) 0;



   /* We write the infoheaderinformation */
   if (fwrite(&infoheader,sizeof(INFOHEADER),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }

   size=width_*height_;

   //printf("size=%d 3*size=%d \n",(int) size, (int) 3*size);

   /* WE STORE THE IMAGE INFORMATION IN image */
   image=(unsigned char*)malloc((size+3)*sizeof(unsigned char));

   long lim[4]={0,size/3,2*size/3,size};

   //printf("size*3=%d\n",size*3);
   for(int k=0;k<3;k++){
     cont=0;
     for(m=lim[k];m<lim[k+1];m++){
       image[cont]=RGB_[m].R_;
       image[cont+1]=RGB_[m].G_;
       image[cont+2]=RGB_[m].B_;
       cont+=3;
     }

     /* We write the image */
       if (fwrite(image,(int) (3*(lim[k+1]-lim[k])),1,fptr) != 1) {
       fprintf(stderr,"Image write failed\n");
       free(image);
       return(-1);
     }
   }

   fclose(fptr);
   free(image);
   return(0);

}


/********************************************************************/
/* APPROXIMATION TO GAUSSIAN FILTER USING ALVAREZ-MAZORRA ALGORITHM */
/********************************************************************/
vector<float> ami_gauss_conv_alma(
vector<float> &rim2,  /* Original picture and output picture*/
float sigma_x,    /* standard deviation of the gaussian in the horizontal axis direction*/
float sigma_y,    /* standard deviation of the gaussian in the vertical axis direction */
float precision,  /* non-negative number to indicate the precisin of the convolution */
int width,int height)
{
   int m,M,cont;
   int i,j,nx1,ny1;

   vector<float> rim=rim2;

   vector<double> z;
   double l_x,v_x,l_y,v_y;
   int nx=height;
   int ny=width;
   int Nc_x=(int) (precision*sigma_x>1?precision*sigma_x:1);
   double t_x=sigma_x*sigma_x/(2*Nc_x);
   int Nc_y=(int) (precision*sigma_y>1?precision*sigma_y:1);
   double t_y=sigma_y*sigma_y/(2*Nc_y);
   int Nc_a=(Nc_x>Nc_y)?Nc_x:Nc_y;

   printf("Nc_a=%d\n",Nc_a);

   /* DEFINIMOS LOS PARAMETROS DEL ALGORITMO EN X E Y*/
   if(t_x>0){
     l_x=(1.+2.*t_x-sqrt((float) 4*t_x+1))/(2*t_x);
     v_x=l_x/t_x;
   }
   if(t_y>0){
     l_y=(1.+2.*t_y-sqrt((float) 4*t_y+1))/(2*t_y);
     v_y=l_y/t_y;
   }
   nx1=nx-1; ny1=ny-1;
   M=nx*ny;

   /* COGEMOS MEMORIA PARA EL VECTOR DE PASO */
   if(nx>ny)
	z.resize(nx);
   else
 	z.resize(ny);

   /* BUCLE PRINCIPAL DE LAS ITERACIONES */
   for(cont=0;cont<Nc_a;cont++){
     if(t_x>0 && cont<Nc_x){
       for(j=0;j<ny;j++){
         z[0]=rim[j]/(1-l_x);
         m=j;
         for(i=1;i<nx;i++){
           z[i]=rim[m+=ny]+l_x*z[i-1];
         }
         rim[m]=z[nx1]/(1-l_x);
         i=nx-1;
         while(i>0){
	       rim[m-ny]=(z[--i]+l_x*rim[m]);
           m-=ny;
         }
       }
	   m=0;
       while(m<M) rim[m++]*=v_x;
	 }
	 if(t_y>0 && cont<Nc_y){
       m=0;
       for(i=0;i<nx;i++){
         z[0]=rim[m]/(1-l_y);
         for(j=1;j<ny;j++){
           z[j]=rim[++m]+l_y*z[j-1];
         }
         rim[m]=z[ny1]/(1-l_y);
         j=ny1;
         while(j>0){
           --m;
	       rim[m]=z[--j]+l_y*rim[m+1];
         }
         m+=ny;
       }
       m=0;
       while(m<M) rim[m++]*=v_y;
     }
  }
  return(rim);
}

/******************************************************************************/
/*           FUNCTION TO COMPUTE HARRIS CORNER QUALITY VALUE                  */
/******************************************************************************/
vector<float> ami_harris_value_estimation(
imageBMP &I, /* ORIGINAL IMAGE */
float sigma) /* STANDARD DEVIATION FOR THE GAUSSIAN */
{
  int width=I.width();
  int height=I.height();
  int i,j;
  float ux,uy,c1,d1,coef1,coef2;
  int m,size=width*height;

  vector<float> harris_value(size); /* HARRIS QUALITY FACTOR */

  coef1=sqrt((float) 2.);
  coef2=0.25*(2.-coef1);
  coef1=0.5*(coef1-1);
  vector<float> image_x(size,0);
  vector<float> image_y(size,0);
  vector<float> image_xy(size,0);

  for(i=0;i<height;i++){
    for(j=0;j<width;j++){
      m=i*width+j;
      if(i==0 || i==(height-1) || j==0 || j==(width-1)){
	    ux=uy=0;
      }
      else{
	    c1= (float) I[m+width+1].R_-I[m-width-1].R_;
	    d1= (float) I[m-width+1].R_-I[m+width-1].R_;
	    uy=coef1*((float) I[m+width].R_-I[m-width].R_)+coef2*(c1-d1);
	    ux=-(coef1*( (float) I[m+1].R_-I[m-1].R_)+coef2*(c1+d1));

      }
      image_x[m]=ux*ux;
      image_y[m]=uy*uy;
      image_xy[m]=ux*uy;
    }
  }

  image_x=ami_gauss_conv_alma(image_x,sigma,sigma,5.,width,height);
  image_y=ami_gauss_conv_alma(image_y,sigma,sigma,5.,width,height);
  image_xy=ami_gauss_conv_alma(image_xy,sigma,sigma,5.,width,height);

  for(int m=0;m<size;m++){
      harris_value[m]=fabs(image_x[m]*image_y[m]-image_xy[m]*image_xy[m]-
	  0.04*(image_x[m]+image_y[m])*(image_x[m]+image_y[m]));
  }

  return(harris_value);
}


imageBMP imageBMP::symmetry(int number_of_sectors,int initial_sector,float center_x,float center_y,float angle0)
{
  float ami_PI=3.141617;
  imageBMP I2;
  int width,height;
  if(number_of_sectors==1){
    number_of_sectors=8;
    width=2*width_;
    height=2*height_;
    center_x=width_;
    center_y=height_;
    I2=imageBMP(width,height,color((unsigned char) 255,(unsigned char) 255,(unsigned char) 255));
    for(int x=0;x<width_;x++){
      for(int y=0;y<height_;y++){
        I2[(y+height_)*width+x+width_]=RGB_[y*width_+x];
        I2[(y+height_)*width+x]=RGB_[y*width_+x];
        I2[(y)*width+x+width_]=RGB_[y*width_+x];
        I2[(y)*width+x]=RGB_[y*width_+x];
      }
    }
  }
  else{
    I2=(*this);
    width=width_;
    height=height_;
  }



  for(int k=1;k<number_of_sectors;k++){
    printf("k=%d\n",k);
    float angle1=angle0+2.*ami_PI*(initial_sector+k)/number_of_sectors;
    float angle2=angle0+2.*ami_PI*(initial_sector+k+1)/number_of_sectors;
    printf("angle1=%lf\n angle2=%lf\n",angle1,angle2);
    float a1=sin(angle1); float b1=-cos(angle1); float c1=-a1*center_x-b1*center_y;
    float a2=sin(angle2); float b2=-cos(angle2); float c2=-a2*center_x-b2*center_y;
    for(int x=0;x<width;x++){
      for(int y=0;y<height;y++){
        if(a1*x+b1*y+c1<=0.1  && a2*x+b2*y+c2>=-0.1){
            int x0=round(x-2*a1*(a1*x+b1*y+c1));
            int y0=round(y-2*b1*(a1*x+b1*y+c1));
            if(0<=x0 && x0<width && 0<=y0 && y0<height)
              I2[x+y*width]=I2[x0+y0*width];
        }
      }
    }
  }

  return(I2);
}

imageBMP imageBMP::symmetry45()
{
  float ami_PI=3.141617;
  imageBMP I2=(*this);
  float angle0=0;
  float center_x=height_/2,center_y=height_/2;
  int initial_sector=2;

  int number_of_sectors=8;
  for(int k=1;k<4;k++){
    //printf("k=%d\n",k);
    float angle1=angle0+2.*ami_PI*(initial_sector+k)/number_of_sectors;
    float angle2=angle0+2.*ami_PI*(initial_sector+k+1)/number_of_sectors;
    //printf("angle1=%lf\n angle2=%lf\n",angle1,angle2);
    float a1=sin(angle1); float b1=-cos(angle1); float c1=-a1*center_x-b1*center_y;
    float a2=sin(angle2); float b2=-cos(angle2); float c2=-a2*center_x-b2*center_y;
    for(int x=0;x<width_;x++){
      for(int y=0;y<height_;y++){
        if(a1*x+b1*y+c1<=0.1  && a2*x+b2*y+c2>=-0.1){
            int x0=round(x-2*a1*(a1*x+b1*y+c1));
            int y0=round(y-2*b1*(a1*x+b1*y+c1));
            if(0<=x0 && x0<width_ && 0<=y0 && y0<height_)
              I2[x+y*width_]=I2[x0+y0*width_];
        }
      }
    }
  }

  for(int x=height_/2;x<width_;x++){
    for(int y=0;y<height_/2;y++){
      I2[x+y*width_]=I2[x+(height_-1-y)*width_];
    }
  }

   for(int x=width_/2;x<width_;x++){
    for(int y=0;y<height_;y++){
      I2[x+y*width_]=I2[width_-1-x+y*width_];
    }
  }

  return(I2);
}

void ami_date_name(char name[200])
{
  time_t timer;
  int m,j;
  char name2[200];
  time(&timer);
  m=((localtime(&timer))->tm_year+1900)*10000;
  m+=((localtime(&timer))->tm_mon+1)*100;
  m+=(localtime(&timer))->tm_mday;
  sprintf(name,"%d",m);

  m=1000000;
  m+=((localtime(&timer))->tm_hour)*10000;
  m+=((localtime(&timer))->tm_min)*100;
  m+=(localtime(&timer))->tm_sec;

  j=100000;
  while(j>=1){
	sprintf(name2,"%d",((int) ((int)(m/j))%10));
	j/=10;
    strcat(name,name2);
  }
  printf("name=%s\n",name);
}

double extract_zoom(char name[200]){
  int pos=-1;
  for(int k=strlen(name)-5;k>=0;k--){
    if(name[k]=='_' && name[k+1]=='z'){
      pos=k+2;
      break;
    }
  }
  if(pos==-1.) return -1.;
  char number[200];
  int n=0;
  for(int k=pos;k<(int) strlen(name)-4;k++) number[n++]=name[k];
  number[n]='\0';
  printf("number string= %s  strlen(number)=%d\n",number,(int) strlen(number));
  for(int k=0;k<(int) strlen(number);k++){
    if(isdigit(number[k])==0 && number[k]!='.') return -1.;
  }
  //printf("atof(number)=%lf\n",atof(number));
  //system("pause");
  return(atof(number));


}

vector < vector<int> > MorphoNeighborhood2D(int width){
  vector < vector<int> > MN;
  int k=0;

  // 1 0
  MN.resize(k+1);
  MN[k].push_back(1);
  MN[k++].push_back(-1);

  // 0 1
  MN.resize(k+1);
  MN[k].push_back(width);
  MN[k++].push_back(-width);

  // 1 1
  MN.resize(k+1);
  MN[k].push_back(1+width);
  MN[k++].push_back(-1-width);

  // 1 -1
  MN.resize(k+1);
  MN[k].push_back(1-width);
  MN[k++].push_back(-1+width);

  return(MN);

}

void morphological_dilation2D(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
 // vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

    // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        I[m]=max(I2[m],I2[m+1],I2[m-1]);
        color c=max(I2[m],I2[m+width],I2[m-width]);
        I[m]=min(I[m],c);
        c=max(I2[m],I2[m+width+1],I2[m-width-1]);
        I[m]=min(I[m],c);
        c=max(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=min(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}

void morphological_erosion2D(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        I[m]=min(I2[m],I2[m+1],I2[m-1]);
        color c=min(I2[m],I2[m+width],I2[m-width]);
        I[m]=max(I[m],c);
        c=min(I2[m],I2[m+width+1],I2[m-width-1]);
        I[m]=max(I[m],c);
        c=min(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=max(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}

/// 2D MEDIAN FILTER BY ITERATION ON A 3X3 NEIGHBORHOOD
void morphological_median2D(imageBMP &I,int Niter){
  printf("init morphological_median2D()...");
  /// WE PERFORM ITERATIONS OF THE MORPHOLOGICAL FILTER
  for(int k=0;k<Niter;k++){
    imageBMP I2=I;

    ///WE COMPUTE MORPHOLOGICAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
//    #pragma omp parallel for
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int iV[9],fV[9];
        int m=m0+x;
        int k=0;
        iV[k]=m; fV[k]=I2[m].R_+I2[m].G_+I2[m].B_; k++;
        iV[k]=m+1; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m-1; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m+width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m-width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m+1+width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m-1+width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m+1-width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_; k++;
        iV[k]=m-1-width; fV[k]=I2[iV[k]].R_+I2[iV[k]].G_+I2[iV[k]].B_;
        /// WE ORDER THE VECTOR
        for(k=0;k<6;k++){
          for(int n=k+1;n<9;n++){
            if( fV[k] > fV[n] ){
               int temp=fV[k]; fV[k]=fV[n]; fV[n]=temp;
               temp=iV[k]; iV[k]=iV[n]; iV[n]=temp;
            }
          }
        }
        I[m]=I2[iV[5]];
      }
    }

  }
  printf("end morphological_median2D()\n");
}

void morphological_erosion2DLineRemoval(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        color c1=min(I2[m],I2[m+1],I2[m-1]);
        color c2=min(I2[m],I2[m+width],I2[m-width]);
        color c3=min(I2[m],I2[m+width+1],I2[m-width-1]);
        color c4=min(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=max(c1,c2);
        color c=max(c1,c3);
        I[m]=min(I[m],c);
        c=max(c1,c4);
        I[m]=min(I[m],c);
        c=max(c2,c3);
        I[m]=min(I[m],c);
        c=max(c2,c4);
        I[m]=min(I[m],c);
        c=max(c3,c4);
        I[m]=min(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}

void morphological_dilation2DLineRemoval(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        color c1=max(I2[m],I2[m+1],I2[m-1]);
        color c2=max(I2[m],I2[m+width],I2[m-width]);
        color c3=max(I2[m],I2[m+width+1],I2[m-width-1]);
        color c4=max(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=min(c1,c2);
        color c=min(c1,c3);
        I[m]=max(I[m],c);
        c=min(c1,c4);
        I[m]=max(I[m],c);
        c=min(c2,c3);
        I[m]=max(I[m],c);
        c=min(c2,c4);
        I[m]=max(I[m],c);
        c=min(c3,c4);
        I[m]=max(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}

//AAA
//void morphological_erosion2DLineRemoval(imageBMP &I,int Niter){
////  printf("init morphological_dilation2D() ... ");
//  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());
//
//  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
//  for(int k=0;k<Niter;k++){
//    // printf("k=%d\n",k);
//    imageBMP I2=I;
//
//    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
//    int height_1=I.height()-1;
//    int width_1=I.width()-1;
//    int width=I.width();
//    for(int y=1;y<height_1;y++){
//      int m0=y*width;
//      for(int x=1;x<width_1;x++){
//        int m=m0+x;
//        I[m]=min(I[m],I2[m+1]);
//        I[m]=min(I[m],I2[m-1]);
//        I[m]=min(I[m],I2[m+width]);
//        I[m]=min(I[m],I2[m-width]);
//        I[m]=min(I[m],I2[m+1+width]);
//        I[m]=min(I[m],I2[m-1+width]);
//        I[m]=min(I[m],I2[m+1-width]);
//        I[m]=min(I[m],I2[m-1-width]);
//      }
//    }
//  }
////  printf("end morphological_dilation2D()\n");
//}
//
//void morphological_dilation2DLineRemoval(imageBMP &I,int Niter){
////  printf("init morphological_dilation2D() ... ");
//  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());
//
//  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
//  for(int k=0;k<Niter;k++){
//    // printf("k=%d\n",k);
//    imageBMP I2=I;
//
//    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
//    int height_1=I.height()-1;
//    int width_1=I.width()-1;
//    int width=I.width();
//    for(int y=1;y<height_1;y++){
//      int m0=y*width;
//      for(int x=1;x<width_1;x++){
//        int m=m0+x;
//        I[m]=max(I[m],I2[m+1]);
//        I[m]=max(I[m],I2[m-1]);
//        I[m]=max(I[m],I2[m+width]);
//        I[m]=max(I[m],I2[m-width]);
//        I[m]=max(I[m],I2[m+1+width]);
//        I[m]=max(I[m],I2[m-1+width]);
//        I[m]=max(I[m],I2[m+1-width]);
//        I[m]=max(I[m],I2[m-1-width]);
//      }
//    }
//  }
////  printf("end morphological_dilation2D()\n");
//}

imageBMP imageBMP::joint(imageBMP &I){
  if(height_==I.height()){
    imageBMP JointImage(width_+I.width(),height_);
    for(int y=0;y<height_;y++){
      for(int x=0;x<width_;x++){
        JointImage[y*JointImage.width()+x]=RGB_[y*width_+x];
      }
      for(int x=0;x<I.width();x++){
        JointImage[y*JointImage.width()+x+width_]=I[y*I.width()+x];
      }
    }
    return(JointImage);
  }

  int heightMax,heightMin;
  if(height_<I.height()){heightMax=I.height(); heightMin=height_;}
  else{ heightMax=height_; heightMin=I.height(); }
  imageBMP JointImage(width_+I.width(),heightMax,color((unsigned char)0,(unsigned char)0,(unsigned char)0));
  for(int y=0;y<heightMin;y++){
    for(int x=0;x<width_;x++){
      JointImage[y*JointImage.width()+x]=RGB_[y*width_+x];
    }
    for(int x=0;x<I.width();x++){
      JointImage[y*JointImage.width()+x+width_]=I[y*I.width()+x];
    }
  }
  return(JointImage);
}

imageBMP imageBMP::jointV(imageBMP &I){
  if(width_!=I.width()) return imageBMP();

  int size2d=width_*height_;
  imageBMP JointImage(width_,height_+I.height());
  for(int y=0;y<height_;y++){
    for(int x=0;x<width_;x++){
      JointImage[y*width_+x]=RGB_[y*width_+x];
    }
    for(int x=0;x<I.width();x++){
      JointImage[y*JointImage.width()+size2d+x]=I[y*width_+x];
    }
  }
  return(JointImage);
}

imageBMP imageBMP::zoom(const float &zoom){
  int widthN=round(width_*zoom);
  int heightN=round(height_*zoom);
  //printf("widthN=%d,heightN=%d\n",widthN,heightN);
  if(widthN<2 || heightN<2) return imageBMP();
  imageBMP iZ(widthN,heightN);
  for(int x=0;x<widthN;x++){
    for(int y=0;y<heightN;y++){
      float xf=x/zoom;
      float yf=y/zoom;
      int xi=xf;
      int yi=yf;
      if(xi>=width_ || yi>=height_){
        iZ[y*widthN+x]=color((unsigned char) 0,(unsigned char) 0,(unsigned char) 0);
        continue;
      }
      float xd=xf-xi;
      float yd=yf-yi;
      int xi1 = xi<(width_-1)?xi+1:xi;
      int yi1 = yi<(height_-1)?yi+1:yi;
      int l00=yi*width_+xi;
      int l10=yi*width_+xi1;
      int l01=yi1*width_+xi;
      int l11=yi1*width_+xi1;
      float c[3];
      //printf("xi=%d,y1=%d,xi1=%d,yi1=%d\n",xi,y1,xi1,yi1);
      c[0]=(float) (1.-xd)*(1.-yd)*RGB_[l00].R_+xd*(1.-yd)*RGB_[l10].R_+(1.-xd)*yd*RGB_[l01].R_+xd*yd*RGB_[l11].R_;
      c[1]=(float) (1.-xd)*(1.-yd)*RGB_[l00].G_+xd*(1.-yd)*RGB_[l10].G_+(1.-xd)*yd*RGB_[l01].G_+xd*yd*RGB_[l11].G_;
      c[2]=(float) (1.-xd)*(1.-yd)*RGB_[l00].B_+xd*(1.-yd)*RGB_[l10].B_+(1.-xd)*yd*RGB_[l01].B_+xd*yd*RGB_[l11].B_;
//      for(int k=0;k<3;k++){
//        c[k]=(float) (1.-xd)*(1.-yd)*RGB_[l00][k]+xd*(1.-yd)*RGB_[l10][k]+(1.-xd)*yd*RGB_[l01][k]+xd*yd*RGB_[l11][k];
//      }
      iZ[y*widthN+x]=color((unsigned char) c[0],(unsigned char) c[1],(unsigned char) c[2]);
    }
  }
  return iZ;
}

vector<float> zoom(vector<float> &If,int &width,int &height,float zoom){
  int widthN=width*zoom;
  int heightN=height*zoom;
  //printf("widthN=%d,heightN=%d\n",widthN,heightN);
  if(widthN<2 || heightN<2){
    width=0;
    height=0;
    return vector<float>();
  }
  vector<float> iZ(widthN*heightN);
  #pragma omp parallel for
  for(int x=0;x<widthN;x++){
    for(int y=0;y<heightN;y++){
      float xf=x/zoom;
      float yf=y/zoom;
      int xi=xf;
      int yi=yf;
      if(xi>=width || yi>=height){
        iZ[y*widthN+x]=0.;
        continue;
      }
      float xd=xf-xi;
      float yd=yf-yi;
      int xi1 = xi<(width-1)?xi+1:xi;
      int yi1 = yi<(height-1)?yi+1:yi;
      int l00=yi*width+xi;
      int l10=yi*width+xi1;
      int l01=yi1*width+xi;
      int l11=yi1*width+xi1;
      //printf("xi=%d,y1=%d,xi1=%d,yi1=%d\n",xi,y1,xi1,yi1);
      iZ[y*widthN+x]=(1.-xd)*(1.-yd)*If[l00]+xd*(1.-yd)*If[l10]+(1.-xd)*yd*If[l01]+xd*yd*If[l11];
    }
  }
  width=widthN;
  height=heightN;
  return iZ;

}

///printf("WE ZOOM, BINARIZE AND FILTER INTENSITY IMAGE\n");
float zoom(vector<float> &If,int &width,int &height,float MinSqArea,float MaxSqArea,float sigma,float BinarizationSlope){
  float result=1.;;
  float sqArea=sqrt((float) width*height);
  if(sqArea>MaxSqArea){
    If=zoom(If,width,height,MaxSqArea/sqArea);
    sqArea=sqrt((float) width*height);
    result=sqArea/MinSqArea;
  }
  else if(sqArea<MinSqArea){
    If=zoom(If,width,height,MinSqArea/sqArea);
    sqArea=sqrt((float) width*height);
  }
  else{
    result=sqArea/MinSqArea;
  }

  If=ami_gauss_conv_alma(If,sigma,sigma,3.,width,height);
  for(int k=0;k<(int) If.size();k++){
    If[k]=127.5+BinarizationSlope*(If[k]-127.5);
    if(If[k]<0.) If[k]=0.;
    else if(If[k]>255.) If[k]=255.;
  }
  return result;

}

/// SORT RANDOMLY
imageBMP imageBMP::sort(){
  double RAND_MAX2b=(double) RAND_MAX*RAND_MAX;
  imageBMP I2=*this;
  for(int k=0;k<(int) RGB_.size();k++){
    int m=((double) rand()/RAND_MAX+rand()/RAND_MAX2b)*RGB_.size();
    m=m<(int) RGB_.size()?m:RGB_.size()-1;
    color c0=I2.RGB_[m];
    I2.RGB_[m]=I2.RGB_[k];
    I2.RGB_[k]=c0;
  }
  return I2;
}

/// SORT HSV
imageBMP imageBMP::sortHSV(){
  imageBMP I2(width_,height_);
  vector<float> measure(RGB_.size());
  vector<int> index(RGB_.size());
  for(int k=0;k<(int) RGB_.size();k++){
    float H,S,V;
    RGB_[k].rgb2hsv(H,S,V);
    measure[k]=4.*H+2.*S+V;
    index[k]=k;
  }

  for(int k=0;k<(int) RGB_.size();k++){
    for(int m=k+1;m<(int) RGB_.size();m++){
      if(measure[k]<measure[m]){
        float t0=measure[m];
        measure[m]=measure[k];
        measure[k]=t0;
        int i0=index[m];
        index[m]=index[k];
        index[k]=i0;
      }
    }
  }

  for(int k=0;k<(int) RGB_.size();k++){
    I2.RGB_[k]=RGB_[index[k]];
  }

  return I2;
}


/// GET A ROI FROM THE IMAGE
imageBMP imageBMP::roi(int x0,int y0,int x1,int y1){
  if( x1>width_ ||  y1>height_ || x0>=x1 || y0>=y1) return imageBMP();

  unsigned int width=x1-x0;
  unsigned int height=y1-y0;

  imageBMP image2(width,height);
  for(int i=y0;i<y1;++i){
    int i2_=i*width_;
    int i2=(i-y0)*width-x0;
    for(int j=x0;j<x1;++j){
      image2[i2+j]=RGB_[i2_+j]  ;
    }
  }
  return image2;

}

/// GENERATION OF A VIDEO BY ZOOMING
void video_generation(imageBMP I,int xc,int yc,int FinalWidth,int Nframes){

  //FinalWidth=2*(FinalWidth/2);
  FinalWidth=3049;
  int FinalHeight=2160;

//  int FinalHeight= round((double) FinalWidth*I.height()/I.width());
//  FinalHeight=2*(FinalHeight/2);

  printf("FinalWidth=%d FinalHeight=%d\n",FinalWidth,FinalHeight);

  if(FinalWidth>=I.width()-1 ||  FinalHeight>=I.height()-1) return;

  printf("Converting image channels to float\n");
  vector<float> Ir(I.size()),Ig(I.size()),Ib(I.size());
  #pragma omp parallel for
  for(int k=0;k<I.size();k++){
     Ir[k]=(float) I[k].R_;
     Ig[k]=(float) I[k].G_;
     Ib[k]=(float) I[k].B_;
  }

  int x0=xc-FinalWidth/4;
  int x1=xc+FinalWidth/4;
  int y0=yc-FinalHeight/4;
  int y1=yc+FinalHeight/4;

  if(x0<=0 || x1>=I.width() || y0<=0 || y1>=I.height() || Nframes<2) return;

  double w0=sqrt((double) Nframes/(Nframes-1));
  double scale=2160./I.height();
  float sigma0=0.;
  for(int k=Nframes;k>=0;k--){
    printf("video frame %d\n",k);
    double w=sqrt((double) k/(Nframes-1));
    float sigma1=0.7*(w0-w)/scale;
    float sigma=sqrt(sigma1*sigma1-sigma0*sigma0);
    if(sigma>0.25){
      sigma0=sigma1;
      printf("sigma=%lf\n",sigma);
      #pragma omp parallel for
      for(int m=0;m<3;m++){
        if(m==0){
          printf("convolution channel red \n");
          Ir=ami_gauss_conv_alma(Ir,sigma,sigma,1,I.width(),I.height());
        }
        if(m==1){
          printf("convolution channel green \n");
          Ig=ami_gauss_conv_alma(Ig,sigma,sigma,1,I.width(),I.height());
        }
        if(m==2){
          printf("convolution channel blue \n");
          Ib=ami_gauss_conv_alma(Ib,sigma,sigma,1,I.width(),I.height());
        }
      }
    }
    //w0=w;

    int x2= w*x0;
    int y2= w*y0;
    int x3= w*x1+(1.-w)*I.width();
    int y3= w*y1+(1.-w)*I.height();

    unsigned int width=x3-x2;
    unsigned int height=y3-y2;

    vector<float> Ir_roi(width*height);
    vector<float> Ig_roi(width*height);
    vector<float> Ib_roi(width*height);

    #pragma omp parallel for
    for(int i=y2;i<y3;++i){
      int i2_=i*I.width();
      int i2=(i-y2)*width-x2;
      for(int j=x2;j<x3;++j){
        Ir_roi[i2+j]=Ir[i2_+j];
        Ig_roi[i2+j]=Ig[i2_+j];
        Ib_roi[i2+j]=Ib[i2_+j];
      }
    }


    int widthNew=width;
    int heightNew=height;
    float zoom2=(double) FinalHeight/height;

    printf("zooming channel red \n");
    vector<float> Ir_zoom=zoom(Ir_roi,widthNew,heightNew,zoom2);
    printf("zooming channel green \n");
    widthNew=width; heightNew=height;
    vector<float> Ig_zoom=zoom(Ig_roi,widthNew,heightNew,zoom2);
    widthNew=width; heightNew=height;
    printf("zooming channel blue \n");
    vector<float> Ib_zoom=zoom(Ib_roi,widthNew,heightNew,zoom2);
    printf("widthNew=%d,heightNew=%d",widthNew,heightNew);

    printf("Creating and saving PNG zoom image \n");
    imageBMP Izoom(widthNew,heightNew);
    #pragma omp parallel for
    for(int k=0;k<Izoom.size();k++){
     Izoom[k].R_=(unsigned char) Ir_zoom[k];
     Izoom[k].G_=(unsigned char) Ig_zoom[k];
     Izoom[k].B_=(unsigned char) Ib_zoom[k];
   }
   Izoom=Izoom.width_extension(3840);
   char name[200];
   sprintf(name,"%d.png",1000+k);
   Izoom.write(name);
   Izoom=Izoom.zoom(0.5);
   sprintf(name,"%d.png",20000+k);
   Izoom.write(name);
  }

}

///// GENERATION OF A VIDEO BY ZOOMING
//void imageBMP::video_generation(int xc,int yc,int FinalWidth,int Nframes){
//
//  FinalWidth=2*(FinalWidth/2);
//  int FinalHeight= round((double) FinalWidth*height_/width_);
//  FinalHeight=2*(FinalHeight/2);
//
//  if(FinalWidth>=width_-1 ||  FinalHeight>=height_-1) return;
//
//  int x0=xc-FinalWidth/4;
//  int x1=xc+FinalWidth/4;
//  int y0=yc-FinalHeight/4;
//  int y1=yc+FinalHeight/4;
//
//  if(x0<=0 || x1>=width_ || y0<=0 || y1>=height_ || Nframes<2) return;
//
//  for(int k=0;k<Nframes;k++){
//    printf("video frame %d\n",k);
//    double w=sqrt((double) k/(Nframes-1));
//    int x2= w*x0;
//    int y2= w*y0;
//    int x3= w*x1+(1.-w)*width_;
//    int y3= w*y1+(1.-w)*height_;
//    imageBMP I=roi(x2,y2,x3,y3);
//    float zoom=(float) FinalWidth/I.width();
//    I=I.zoom(zoom);
//    char name[200];
//    sprintf(name,"%d.png",1000+k);
//    if (I.height()!=FinalHeight){
//      imageBMP I2(FinalWidth,FinalHeight,color((unsigned char) 0));
//      for(unsigned m=0;m<(unsigned int) I2.size() && m<(unsigned int) I.size();m++) I2[m]=I[m];
//      I2.write(name);
//    }
//    else{
//      I.write(name);
//    }
//  }
//
//}

/// WIDTH EXTENSION
imageBMP imageBMP::width_extension(int newWidth){
  if(width_>=newWidth) return *this;
  imageBMP I(newWidth,height_,color(0));
  //int disp=newWidth-width_;//
  int disp=(newWidth-width_)/2;
  for(int x=0;x<width_;x++){
    for(int y=0;y<height_;y++){
      I[x+disp+y*newWidth]=RGB_[x+y*width_];
    }
  }
  return I;

}

