/**
 * \file color.cpp
 * \brief class to manage color values
 * \author Luis Alvarez \n \n
*/


#include "color.h"
#include "stdio.h"

/// CONVERT HSV TO RGB
void color::hsv2rgb(float h,float s,float v){
  float      hh, p, q, t, ff;
  long        i;
  float      r,g,b;

  if(s <= 0.0) {       // < is bogus, just shuts up warnings
    r = v;
    g = v;
    b = v;
  }
  else{
    hh = h;
    if(hh >= 360.0) hh = 0.0;
    hh /= 60.0;
    i = (long)hh;
    ff = hh - i;
    p = (color_type) (v * (1.0 - s));
    q = (color_type) (v * (1.0 - (s * ff)));
    t = (color_type) (v * (1.0 - (s * (1.0 - ff))));

    switch(i) {
    case 0:
        r = v;
        g = t;
        b = p;
        break;
    case 1:
        r = q;
        g = v;
        b = p;
        break;
    case 2:
        r = p;
        g = v;
        b = t;
        break;

    case 3:
        r = p;
        g = q;
        b = v;
        break;
    case 4:
        r = t;
        g = p;
        b = v;
        break;
    case 5:
    default:
        r = v;
        g = p;
        b = q;
        break;
    }
  }

  //R=round(R*255.); G=round(255.*G); B=round(255.*B);
  R_ = (r<0?0:(r>=255?255:(color_type) r));
  G_ =  (g<0?0:(g>=255?255:(color_type) g));
  B_ =  (b<0?0:(b>=255?255:(color_type) b));
  //if(R_==0 && G_==0 && B_==0){printf("i=%d, h=%f s=%f v=%f\n",i,h,s,v); system("pause"); }
}

// CONVERT RGB TO HSV
void color::rgb2hsv(float &H,float &S,float &V){
  float rgb_min,rgb_max;
  float R=R_,G=G_,B=B_;
  rgb_min=rgb_max=R;
  if(rgb_min>G) rgb_min=G;
  if(rgb_min>B) rgb_min=B;
  if(rgb_max<G) rgb_max=G;
  if(rgb_max<B) rgb_max=B;

  V = rgb_max;                                // v
  float delta = rgb_max - rgb_min;
  if (delta < 0.00001)
  {
     S = 0;
     H = 0; // undefined, maybe nan?
    return;
  }
  if( rgb_max > 0.0 ) { // NOTE: if Max is == 0, this divide would cause a crash
    S = (delta / rgb_max);                  // s
  }
  else {
    S = 0.0;
    H = 0.;                            // its now undefined
    return;
  }
  if( R_ == rgb_max )                           // > is bogus, just keeps compilor happy
    H = ( G - B ) / delta;        // between yellow & magenta
  else if( G == rgb_max )
    H = 2.0 + (  B - R ) / delta;  // between cyan & yellow
  else
    H = 4.0 + (  R - G ) / delta;  // between magenta & cyan

  H *= 60.0;                              // degrees

  if( H < 0.0 )  H += 360.0;
}

color min(color &c1,color &c2){
  if( ( (int) c1.R_-c2.R_+c1.G_-c2.G_+c1.B_-c2.B_) <0 )
    return c1;
  return c2;
}

color max(color &c1,color &c2){
  if( ( (int) c1.R_-c2.R_+c1.G_-c2.G_+c1.B_-c2.B_) <0 )
    return c2;
  return c1;
}

color min(color &c1,color &c2,color &c3){
   if( ( (int) c1.R_-c2.R_+c1.G_-c2.G_+c1.B_-c2.B_) <0 ){
     if( ( (int) c1.R_-c3.R_+c1.G_-c3.G_+c1.B_-c3.B_) <0 ){
        return c1;
     }
     return c3;
   }
   else{
     if( ( (int) c2.R_-c3.R_+c2.G_-c3.G_+c2.B_-c3.B_) <0 ){
        return c2;
     }
     return c3;
   }
}

color max(color &c1,color &c2,color &c3){
   if( ( (int) c1.R_-c2.R_+c1.G_-c2.G_+c1.B_-c2.B_) >0 ){
     if( ( (int) c1.R_-c3.R_+c1.G_-c3.G_+c1.B_-c3.B_) >0 ){
        return c1;
     }
     return c3;
   }
   else{
     if( ( (int) c2.R_-c3.R_+c2.G_-c3.G_+c2.B_-c3.B_) >0 ){
        return c2;
     }
     return c3;
   }
}

float color_ga_1(float u){
  if(u<=0.04045) return u/12.92;
  else return pow( (u+0.055)/1.055 , 2.4);
}

float color_ga(float u){
  if(u<=0.0031308) return u*12.92;
  else return 1.055*pow(u,0.416666666666666666667)-0.055;
}

void RGB2XYZ(float R,float G,float B,float &X,float &Y,float &Z){
  R=color_ga_1(R/255.);
  G=color_ga_1(G/255.);
  B=color_ga_1(B/255.);

  X=0.4124*R+0.3576*G+0.1805*B;
  Y=0.2126*R+0.7152*G+0.0722*B;
  Z=0.0193*R+0.1192*G+0.9505*B;
}

void XYZ2RGB(float X,float Y,float Z,float &R,float &G,float &B){

  R= 3.2406*X-1.5372*Y-0.4986*Z;
  G=-0.9689*X+1.8758*Y+0.0415*Z;
  B= 0.0557*X-0.2040*Y+1.0570*Z;

  R=255.*color_ga(R);
  G=255.*color_ga(G);
  B=255.*color_ga(B);

}

float color_f(float u){
  if(u>8.856451679e-3) return pow(u,1./3.);
  else return u*7.787037037+0.1379310345;
}

float color_f_1(float u){
  if(u>0.2068965517) return u*u*u;
  else return (u-0.1379310345)/7.787037037;
}

void XYZ2Lab(float X,float Y,float Z,float &L,float &a,float &b){

  L=116.*color_f(Y)-16;
  a=500.*(color_f(X/0.950489)-color_f(Y));
  b=200.*(color_f(Y)-color_f(Z/1.08884));
}

void Lab2XYZ(float L,float a,float b,float &X,float &Y,float &Z){

  X = 0.950489 * color_f_1( (L+16)/116.+a/500. );
  Y=  color_f_1( (L+16)/116. );
  Z= 1.08884 * color_f_1( (L+16)/116.-b/200. );
}

void Lab2RGB(float L,float a,float b,float &R,float &G,float &B){
  float X,Y,Z;
  Lab2XYZ(L,a,b,X,Y,Z);
  XYZ2RGB(X,Y,Z,R,G,B);
}

void RGB2Lab(float R,float G,float B,float &L,float &a,float &b){
  float X,Y,Z;
  RGB2XYZ(R,G,B,X,Y,Z);
  XYZ2Lab(X,Y,Z,L,a,b);
}


