/**
 * \file color.h
 * \brief class to manage color values
 * \author Luis Alvarez \n \n
*/


#ifndef COLOR_H_
#define COLOR_H_
#include <iostream>
#include <vector>
#include <stdlib.h>
#include <math.h>

#define color_type float

/**
 * \class  color
 * \brief class  to store color
 * \author Luis Alvarez
 */
class color{
  public :
  color_type R_ /** red color channel */;
  color_type G_ /** green color channel */;
  color_type B_ /** blue color channel */;
  color(unsigned char  R,unsigned char G,unsigned char B){R_=(color_type) R; G_=(color_type)G; B_=(color_type)B;}
  color(color_type c_){R_=c_; G_=c_; B_=c_;}
  color(float H,float S,float V){hsv2rgb(H,S,V);}
  color(){};
  void hsv2rgb(float H,float S,float V);
  void rgb2hsv(float &H,float &S,float &V);
  void random(){
    R_ = (color_type) (255.*rand()/RAND_MAX);
    G_ = (color_type) (255.*rand()/RAND_MAX);
    B_ = (color_type) (255.*rand()/RAND_MAX);
  }
  color operator*(float f){
     float Rf=f*R_;  float Gf=f*G_;  float Bf=f*B_;
     color c;
     c.R_=Rf<0.?0:(Rf>255.?255:(color_type)Rf);
     c.G_=Gf<0.?0:(Gf>255.?255:(color_type)Gf);
     c.B_=Bf<0.?0:(Bf>255.?255:(color_type)Bf);
     return c;
  }
  color operator+(color c2){
     float Rf=(float) R_+c2.R_;
     float Gf=(float) G_+c2.G_;
     float Bf=(float) B_+c2.B_;
     color c;
     c.R_=Rf<0.?0:(Rf>255.?255:(color_type)Rf);
     c.G_=Gf<0.?0:(Gf>255.?255:(color_type)Gf);
     c.B_=Bf<0.?0:(Bf>255.?255:(color_type)Bf);
     return c;
  }
  bool operator==(color c2){
    if(R_==c2.R_ && G_==c2.G_ && B_==c2.B_) return true;
    return false;
  }
};



color min(color &c1,color &c2);
color max(color &c1,color &c2);
color min(color &c1,color &c2,color &c3);
color max(color &c1,color &c2,color &c3);

float color_f(float u);
float color_f_1(float u);

void Lab2RGB(float L,float a,float b,float &R,float &G,float &B);
void RGB2Lab(float R,float G,float B,float &L,float &a,float &b);
void XYZ2Lab(float X,float Y,float Z,float &L,float &a,float &b);
void Lab2XYZ(float L,float a,float b,float &X,float &Y,float &Z);
void RGB2XYZ(float R,float G,float B,float &X,float &Y,float &Z);
void XYZ2RGB(float X,float Y,float Z,float &R,float &G,float &B);

#endif
