/**
 * \file shape.cpp
 * \brief functions  to manage shapes
 * \author Luis Alvarez \n \n
*/


#include <iostream>
#include <vector>
#include <math.h>
#include "shape.h"


/*************************************************************************
FUNCTIONS TO DRAW A SHAPE IN AN IMAGE. IMAGE COORDINATES ARE RESCALED USING PARAMETER scaleI2Tr
*************************************************************************/
void draw_shape(shape &S0,imageBMP &I,float scaleI2Tr,vector<imageBMP> &PaletteDataBase,int OrderToDrawShapes,bool periodicity,int MaxVideoFrames){
  vector < vector < vector<color> > > TessPalette;
  int palette_index=-1;

  shape S;

  unsigned int Ntriangulations=0;
  S0.count_triangulations(Ntriangulations);
  int VideoFrameRate=0;
  if(MaxVideoFrames>0){
    if((int) Ntriangulations<=MaxVideoFrames) VideoFrameRate=1;
    else VideoFrameRate=Ntriangulations/MaxVideoFrames+1;
  }
  printf("Ntriangulations=%d,MaxVideoFrames=%d,VideoFrameRate=%d\n",Ntriangulations,MaxVideoFrames,VideoFrameRate);


  if(OrderToDrawShapes<0){
    S=S0.extract_main_shape();
    for(int m=0;m<(int) S0.shapeChild.size();m++){
      shape S2=S0.shapeChild[m].extract_main_shape();
      S0.shapeChild[m].tree_level_set(0);
      vector <shape*> SpV;
      S0.shapeChild[m].build_shape_pointer_vector(SpV);
      //printf("SpV.size()=%d\n",SpV.size());
      int level=1;
      while(true){
        vector<int> index;
        for(int k=1;k<(int) SpV.size();k++){
          if((*SpV[k]).tree_level==level) index.push_back(k);
        }
        //printf("level=%d index.size()=%d\n",level,index.size());
        if(index.size()==0) break;
        for(int k=0;k<(int) index.size();k++){
          int temp=index[k];
          int j=ami_randI(index.size());
          index[k]=index[j];
          index[j]=temp;
        }
        for(int k=0;k<(int) index.size();k++){
          S2.shapeChild.push_back((*SpV[index[k]]).extract_main_shape());
        }
        level++;
      }
      S.shapeChild.push_back(S2);
    }
  }
  else{
    S=S0;
  }

  triangulation T;
  if(S.T.p.size()>0){
    vector<unsigned char> keepFatherInclusionV(I.size(),0);
    draw_shape(S,I,scaleI2Tr,TessPalette,palette_index,PaletteDataBase,T,keepFatherInclusionV,periodicity,VideoFrameRate);
  }
  else{
    for(int k=0;k<(int) S.shapeChild.size();k++){
      vector<unsigned char> keepFatherInclusionV(I.size(),0);
      draw_shape(S.shapeChild[k],I,scaleI2Tr,TessPalette,palette_index,PaletteDataBase,T,keepFatherInclusionV,periodicity,VideoFrameRate);
    }
  }

}

static unsigned int shape_counter=0;
static vector<imageBMP>iBMPV(24);
static unsigned int shape_drawn_counter=0;

void draw_shape(shape &S,imageBMP &I,float scaleI2Tr,vector < vector < vector<color> > > &TessPalette,int &palette_index,
                vector<imageBMP> &PaletteDataBase,triangulation &T,vector<unsigned char> &keepFatherInclusionV,bool periodicity,int VideoFrameRate)
{

  //printf("INIT draw_shape \n");
  //printf("VideoFrameRate=%d\n",VideoFrameRate);
  //printf("T.p.size()=%d\n",T.p.size());
  if(S.T.p.size()<3) return;
  //printf("S.c=(%lf,%lf,%lf)\n",(float) S.c.R_,(float) S.c.G_,(float) S.c.B_);
  if(S.tesselation_palette_index>=0 && (S.tesselation_palette_index!=palette_index || TessPalette.size()!=256) ){
    //printf("S.tesselation=%d TessPalette.size()=%d\n",S.tesselation_palette_index,TessPalette.size()) ; system("pause");
    palette_index=S.tesselation_palette_index;
    TessPalette=vector < vector < vector<color> > >(256,vector< vector < color > >(256,vector < color >(256)));
    for(int i=0;i<256;i++){
      for(int j=0;j<256;j++){
        for(int k=0;k<256;k++){
          TessPalette[i][j][k]=PaletteDataBase[S.tesselation_palette_index%PaletteDataBase.size()][(int) ((double) PaletteDataBase[S.tesselation_palette_index%PaletteDataBase.size()].size()*rand()/(RAND_MAX+1.))];
        }
      }
    }
    TessPalette[255][255][255]=color((unsigned char) 0,(unsigned char) 0,(unsigned char) 0);
    TessPalette[0][0][0]=color((unsigned char) 255,(unsigned char) 255,(unsigned char) 255);
  }

  int width=I.width();
  int height=I.height();

  if(S.T.n.size()!=S.T.t.size()) S.T.update_n();

   /** IF LIGHT SPOT EFFECT IS USED WE COMPUTE THE SHAPE COLOR HSV */
   float H0,S0,V0;
   vector<float> line;
   float max_line_distance=0;;
   if(S.light_spot>0.){
    S.c.rgb2hsv(H0,S0,V0);
   }
   /// IF BORDER SPOT IS ACTIVATED WE COMPUTE A BORDER LINE
   else if(S.light_spot<0.) { //AAA
      //printf("ok1\n");
      line.resize(3);
      S.c.rgb2hsv(H0,S0,V0);
      if(S.T.n.size()!=S.T.t.size()) S.T.update_n();
      int k0=rand()%((int)S.T.n.size());
      int i0=rand();
      i0=i0%3;
      for(int k=0;k<(int) S.T.n.size();k++){
        int k1=(k+k0)%((int) S.T.n.size());
        for(int i=0;i<3;i++){
          int i1=(i+i0)%3;
          //printf("k1=%d,T.n.size()=%d,i1=%d\n",k1,T.n.size(),i1);
          if(S.T.n[k1].v[i1]<0){
            point2d n1=(S.T.p[S.T.t[k1].v[i1]]-S.T.p[S.T.t[k1].v[(i1+1)%3]]).normal();
            line[0]=n1.x;
            line[1]=n1.y;
            line[2]=-n1.x*S.T.p[S.T.t[k1].v[i1]].x-n1.y*S.T.p[S.T.t[k1].v[i1]].y;
            //printf("(%lf,%lf) (%lf,%lf)\n",S.T.p[S.T.t[k1].v[i1]].x,S.T.p[S.T.t[k1].v[i1]].y,S.T.p[S.T.t[k1].v[(i1+1)%3]].x,S.T.p[S.T.t[k1].v[(i1+1)%3]].y);
            //printf("line= %lf x+ %lf y + %lf=0\n",line[0],line[1],line[2]);
//            system("pause");
            k=S.T.n.size();
            break;
          }
        }
      }
      for(int k=0;k<(int) S.T.p.size();k++){
        float d=fabs(line[0]*S.T.p[k].x+line[1]*S.T.p[k].y+line[2]);
        if(d>max_line_distance) max_line_distance=d;
      }
  }

  /** WE STORE TRIANGULATION POINTS IN IMAGE PIXEL COORDINATES */
  vector<int> pix(S.T.p.size());
  vector<int> piy(S.T.p.size());

  #pragma omp parallel for if(pix.size()>16)
  for(unsigned int k=0;k<pix.size();k++){
    pix[k]=round(S.T.p[k].x/scaleI2Tr);
    piy[k]=round(S.T.p[k].y/scaleI2Tr);
  }
  ///printf("WE COMPUTE THE BOUNDING BOX OF THE SHAPE \n");
  int pix_min=pix[0],pix_max=pix[0],piy_min=piy[0],piy_max=piy[0];
  for(unsigned int k=0;k<pix.size();k++){
    if(pix[k]<pix_min) pix_min=pix[k];
    else if(pix[k]>pix_max) pix_max=pix[k];
    if(piy[k]<piy_min) piy_min=piy[k];
    else if(piy[k]>piy_max) piy_max=piy[k];
  }
  pix_max++; piy_max++; pix_min--; piy_min--;
  if(pix_min<0) pix_min=0;
  if(pix_max>width) pix_max=width;
  if(piy_min<0) piy_min=0;
  if(piy_max>height) piy_max=height;
  if(pix_min>=pix_max || piy_min>=piy_max) return;

  /// WE CONTROL THAT EACH POINT OF THE SHAPE IS ONLY VISITED ONCE
  int pix_width=pix_max-pix_min;
  //printf("pix_min=%d pix_max=%d piy_min=%d piy_max=%d\n",pix_min,pix_max,piy_min,piy_max);
  //printf("visited.size()=%d\n",(pix_max-pix_min)*(piy_max-piy_min)+1);
  vector<bool> visited((pix_max-pix_min)*(piy_max-piy_min)+1,false);
  ///printf("WE GO TROUGH TRIANGLES\n");
  if(periodicity==true){
    #pragma omp parallel for if(S.T.t.size()>16)
    for(unsigned int l=0;l<S.T.t.size();l++){

//S.c.random();

      /// WE STORE IN i,j,k THE INDEX OF TRIANGLE POINTS ORDERED BY y COORDINATES
      int i=S.T.t[l].v[0];
      int j=S.T.t[l].v[1];
      int k=S.T.t[l].v[2];
      int nij=S.T.n[l].v[0];
      int njk=S.T.n[l].v[1];
      int nik=S.T.n[l].v[2];
      if(piy[i]>piy[j]){int temp=i; i=j; j=temp; temp=njk; njk=nik; nik=temp;}
      if(piy[i]>piy[k]){int temp=i; i=k; k=temp; temp=njk; njk=nij; nij=temp;}
      if(piy[i]>=height) continue;
      if(piy[j]>piy[k]){int temp=j; j=k; k=temp; temp=nik; nik=nij; nij=temp;}
      int horizontal_edge=false; // horizontal_edge manages the case coordinate y of points i,j are equal.

      /// WE COMPUTE TRIANGLE MIN y COORDINATE
      int ymax;
      int ymin=piy[i]==0?0:piy[i]+1;

     /// WE MANAGE THE CASE OF HORIZONTAL TRIANGLE EDGE
     if(piy[i]==piy[j]){
        ymax=piy[k];
        horizontal_edge=true;
        int temp=i; i=k; k=temp; // we interchange i and k in order to manage the triangle from the non-horizontal vertex.
        temp=njk; njk=nij; nij=temp;
      }
      else{
        ymax=piy[j];
      }

      /** WE DRAW THE TRIANGLE IN 2 PARTS */
      for(int mm=0;mm<2;mm++){
        ///WE MANAGE FIRST PART OF TRIANGLE
        float p0=(float) (pix[k]-pix[i])/(piy[k]-piy[i]); //slope of line of vertex i-k
        float p1=(float) (pix[j]-pix[i])/(piy[j]-piy[i]); //slope of line of vertex i-j

        ///WE INTERCHANGE SLOPES VALUE IN ORDER xmin<xmax
        if((p1<p0 && horizontal_edge==false) || (p1>p0 && horizontal_edge==true) ){float tempd=p1; p1=p0; p0=tempd;}

        /// WE GO THROUGH FIRST PART TRIANGLE PIXELS
        if(mm==1){
          /// WE FIT y COORDINATE INTERVAL
          ymin=ymax+1;
          ymax=piy[k];
          if(ymin==ymax) continue;
          /// WE INTERCHANGE THE ROLE OF i AND k
          int temp=k; k=i; i=temp;
          temp=njk; njk=nij; nij=temp;
          p0=(float) (pix[k]-pix[i])/(piy[k]-piy[i]);
          p1=(float) (pix[j]-pix[i])/(piy[j]-piy[i]);
          /// WE INTERCHANGE SLOPES VALUE IN ORDER xmin<xmax
          if(p1>p0){float tempd=p1; p1=p0; p0=tempd;}
        }
        #pragma omp parallel for if(ymax-ymin>16)
        for(int y=ymin;y<=ymax;y++){
          /// WE COMPUTE INTERSECTION OF LINES y=y AND LINES i-j and i-k
          int xmin=round(pix[i]+(y-piy[i])*p0);
          int xmax=round(pix[i]+(y-piy[i])*p1)-1;
          /// WE ASSIGN IMAGE COLOR
          int y0=y;
          while(y0<0){ y0+=height; } y0=y0%height;
          int pos=y0*width;
          for(int x=xmin;x<=xmax;x++){
            int x0=x; while(x0<0) x0+=width; x0=x0%width;
            if(keepFatherInclusionV[pos+x0]==255) continue;
            int visited_pos=(x-pix_min)+(y-piy_min)*pix_width;
            if(visited[visited_pos]==true) continue;
            visited[visited_pos]=true;
            if(T.p.size()>0 && S.keepFatherInclusion==0){
              point2d p(x0*scaleI2Tr,y0*scaleI2Tr);
              if(T.inclusion_test(p)<0){
                keepFatherInclusionV[pos+x0]=255;
                continue;
              }
            }
            if(S.tesselation_palette_index>=0){
              int iR=round((float) I[pos+x0].R_+S.c.R_)/2;
              int iG=round((float) I[pos+x0].G_+S.c.G_)/2;
              int iB=round((float) I[pos+x0].B_+S.c.B_)/2;
              I[pos+x0]=TessPalette[iR][iG][iB];
            }
            else if(S.light_spot>0.){
              float norm=(point2d(x,y)*scaleI2Tr-S.light_spot_center).norm();
              float V2=V0*(1./(1.+S.light_spot*norm));
              color c(H0,S0,V2);
              I[pos+x0]= ( c*(1-S.transparency)+I[pos+x0]*S.transparency);
            }
            else{
              I[pos+x0]=(S.c*(1-S.transparency)+I[pos+x0]*S.transparency);
            }
          }
        }
        if(horizontal_edge==true) break; // IF WE DEAL WITH A NON-HORIZONTAL TRIANGLE WE DEAL WITH THE SECOND PART
      }
    }
  }
  else{ /// no periodicity case
    /// WE GO TROUGH TRIANGLES
    #pragma omp parallel for if(S.T.t.size()>16)
    for(unsigned int l=0;l<S.T.t.size();l++){
      //printf("%d %d ",l,(int) S.T.t.size());
//S.c.random();

      /// WE STORE IN i,j,k THE INDEX OF TRIANGLE POINTS ORDERED BY y COORDINATES
      int i=S.T.t[l].v[0],j=S.T.t[l].v[1],k=S.T.t[l].v[2];
      if(piy[i]>piy[j]){int temp=i; i=j; j=temp; }
      if(piy[i]>piy[k]){int temp=i; i=k; k=temp; }
      if(piy[i]>=height) continue;
      if(piy[j]>piy[k]){int temp=j; j=k; k=temp; }
      int horizontal_edge=false; // horizontal_edge manages the case coordinate y of points i,j are equal.

      /// WE COMPUTE TRIANGLE MIN y AND MAX y COORDINATES FOR THE FIRST SECTION OF THE TRIANGLE
      int ymin=piy[i],ymax=piy[j];

      /// WE MANAGE THE CASE OF HORIZONTAL TRIANGLE EDGE
      if(piy[i]==piy[j]){
        ymax=piy[k]; // we compute ymax
        horizontal_edge=true;
        int temp=i; i=k; k=temp; // we interchange i and k in order to manage the triangle from the non-horizontal vertex.
      }

      /** WE DRAW THE TRIANGLE IN 2 PARTS */
      for(int mm=0;mm<2;mm++){
        /// WE MANAGE FIRST PART OF TRIANGLE
        float p0=0.,p1=0.;
        if(ymax>ymin){
          p0=(float) (pix[k]-pix[i])/(piy[k]-piy[i]); //slope of line of vertex i-k
          p1=(float) (pix[j]-pix[i])/(piy[j]-piy[i]); //slope of line of vertex i-j
          /// WE INTERCHANGE SLOPES VALUE IN ORDER xmin<xmax
          if((p1<p0 && horizontal_edge==false) || (p1>p0 && horizontal_edge==true) ){float tempd=p1; p1=p0; p0=tempd;}
        }
        /// WE UPDATE INFORMATION FOR THE SECOND PART OF THE TRIANGLE.
        if(mm==1){
          /// WE FIT y COORDINATE INTERVAL
          ymin=ymax+1;
          ymax=piy[k];
          /// WE INTERCHANGE THE ROLE OF i AND k
          int temp=k; k=i; i=temp;
          if(ymax>ymin){
            p0=(float) (pix[k]-pix[i])/(piy[k]-piy[i]);
            p1=(float) (pix[j]-pix[i])/(piy[j]-piy[i]);
            /// WE INTERCHANGE SLOPES VALUE IN ORDER xmin<xmax
            if(p1>p0){float tempd=p1; p1=p0; p0=tempd;}
          }
        }
        if(ymin<0) {ymin=0;} if(ymax>=height){ ymax=height-1;}

        /// WE GO THROUGH THE HORIZONTAL SEGMENT OF TRIANGLE SECTION
        #pragma omp parallel for if(ymax-ymin>16)
        for(int y=ymin;y<=ymax;y++){
          int xmin,xmax;
          if(ymax==ymin){
            if(mm==0){ /// case of a flat triangle piy[i]==piy[j]==piy[k];
              xmin=pix[i]<pix[j]?pix[i]:pix[j];
              xmax=pix[i]<pix[j]?pix[j]:pix[i];
              if(pix[k]<xmin) xmin=pix[k];
              else if(pix[k]>xmax) xmax=pix[k];
              //xmin--; xmax++;
              //printf("(%d,%d)x(%d,%d)\n",xmin,xmax,ymin,ymax);
            }
            else{ /// case piy[j]+1==piy[i]
             //continue;
             xmin=xmax=pix[i];
            }
          }
          else{
            xmin=round(pix[i]+(y-piy[i])*p0);
            xmax=round(pix[i]+(y-piy[i])*p1);
          }
          if(xmin<0) xmin=0;
          if(xmax>=width) xmax=width-1;

          /// WE ASSIGN IMAGE COLOR
          int pos=y*width;
          for(int x=xmin;x<=xmax;x++){
            if(keepFatherInclusionV[pos+x]==255) continue;
            int visited_pos=(x-pix_min)+(y-piy_min)*pix_width;
//if(visited_pos<0 || visited_pos>= visited.size()){
//  printf("visited_pos=%d, visited.size()=%d\n",visited_pos, visited.size());
//  system("pause");
//}
            if(visited[visited_pos]==true) continue;
            visited[visited_pos]=true;
            if(T.p.size()>0 && S.keepFatherInclusion==0){
              point2d p(x*scaleI2Tr,y*scaleI2Tr);
              if(T.inclusion_test(p)<0){
                keepFatherInclusionV[pos+x]=255; ///AAAA
                continue;
              }
            }
            if(S.tesselation_palette_index>=0){
              int iR=round((float) I[pos+x].R_+S.c.R_)/2;
              int iG=round((float) I[pos+x].G_+S.c.G_)/2;
              int iB=round((float) I[pos+x].B_+S.c.B_)/2;
              I[pos+x]=TessPalette[iR][iG][iB];
            }
            else if(S.light_spot>0.){
              float norm=(point2d(x,y)*scaleI2Tr-S.light_spot_center).norm();
              float V2=V0*(1./(1.+S.light_spot*norm));
              color c(H0,S0,V2);
              I[pos+x]=c*(1-S.transparency)+I[pos+x]*S.transparency;
            }
            else if(S.light_spot<0. && line.size()==3){
              point2d p2d=point2d(x,y)*scaleI2Tr;
              float d=1.-fabs(line[0]*p2d.x+line[1]*p2d.y+line[2])/max_line_distance; ///AAA
              d*=d;
              d*=d;
              float V2=V0*(1./(1.-S.light_spot*d));
              color c(H0,S0,V2);
              I[pos+x]=c*(1-S.transparency)+I[pos+x]*S.transparency;
            }
            else{
              I[pos+x]=S.c*(1-S.transparency)+I[pos+x]*S.transparency;
            }
          }
        }
        if(horizontal_edge==true) break; // IF WE DEAL WITH A NON-HORIZONTAL TRIANGLE WE DEAL WITH THE SECOND PART
      }
    }

  }

  /// WE SAVE A VIDEO SEQUENCE ///AAA
  if(VideoFrameRate>0 && S.T.t.size()>0 && (S.c.R_!=0 || S.c.G_!=0 || S.c.B_!=0) ){
    printf("shape_drawn_counter=%d\n",shape_drawn_counter);
    iBMPV[shape_drawn_counter%24]=S.T.draw(260,6,color((unsigned char)0),color((unsigned char)255));
    shape_drawn_counter++;
  }
  if(VideoFrameRate>0 && shape_counter%VideoFrameRate==0){
    char filename[200];
    sprintf(filename,"%d.png",10000+shape_counter/VideoFrameRate);
    if(I.height()!=2160){
      I.write(filename);
    }
    else{
      imageBMP I2=I.width_extension(3840);
      int x=0;
      for(int m=0;m<24;m++){
        int y=m%8;
        for(int i=0;i<iBMPV[m].width();i++){
          for(int j=0;j<iBMPV[m].height();j++){
            I2[ (j+y*270)*I2.width()+i+x*263]=iBMPV[m][j*iBMPV[m].width()+i];
          }
        }
        if(y==7) x++;
      }
      I2.write(filename);
      sprintf(filename,"%d.png",20000+shape_counter/VideoFrameRate);
      imageBMP I3=I2.zoom(0.5);
      I3.write(filename);
    }
  }
  shape_counter++;

  ///printf("WE DRAW SHAPE BORDER\n");
  for(int k=0;k<(int) S.shapeBorder.size();k++){
    S.shapeBorder[k].keepFatherInclusion=S.keepFatherInclusion;
    draw_shape(S.shapeBorder[k],I,scaleI2Tr,TessPalette,palette_index,PaletteDataBase,T,keepFatherInclusionV,periodicity,VideoFrameRate);
  }
  ///printf("WE DRAW SHAPE CHILDREN\n");
  for(int k=0;k<(int) S.shapeChild.size();k++){
    draw_shape(S.shapeChild[k],I,scaleI2Tr,TessPalette,palette_index,PaletteDataBase,S.T,keepFatherInclusionV,periodicity,VideoFrameRate);
  }
  //printf("end draw_shape()\n");

}


/** UPDATING EXTREMA USING ALL SHAPE TREE BOUNDARY BOX*/
void shape::extrema_update()
{
  if(T.p.size()>0) extrema=T.extrema();
  int k2=-1;
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].extrema_update();
    if(shapeChild[k].extrema.size()==2) k2=k;
  }
  if(T.p.size()<2 && k2>=0) extrema=shapeChild[k2].extrema;
  for(int k=0;k<(int) shapeChild.size();k++){
    if(shapeChild[k].extrema.size()==2){
      if(extrema[0].x>shapeChild[k].extrema[0].x) extrema[0].x=shapeChild[k].extrema[0].x;
      if(extrema[0].y>shapeChild[k].extrema[0].y) extrema[0].y=shapeChild[k].extrema[0].y;
      if(extrema[1].x<shapeChild[k].extrema[1].x) extrema[1].x=shapeChild[k].extrema[1].x;
      if(extrema[1].y<shapeChild[k].extrema[1].y) extrema[1].y=shapeChild[k].extrema[1].y;
    }
  }
}

/** a basic print of a shape */
void shape::print()
{
  if(extrema.size()==2){
    printf("level : %d, extrema=(%1.2lf,%1.2lf), (%1.2lf,%1.2lf) ",tree_level,extrema[0].x,extrema[0].y,extrema[1].x,extrema[1].y);
    printf("color=(%d,%d,%d)\n",(int) c.R_,(int) c.G_,(int) c.B_);
  }

  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].print();
  }

}

///* FUNCTION TO TEST IF 2 SEGMENTS INTERCEPT (RETURN 0) IF THE SEGMENTS INTERCEPT AND RETURN -1 OTHERWISE */
int ami_test_segment_interception(float x0,float y0,float x1,float y1,float x2,float y2,float x3,float y3)
{
  float a=(x2-x0)*(y1-y0)-(y2-y0)*(x1-x0);
  float b=(x3-x0)*(y1-y0)-(y3-y0)*(x1-x0);
  if(a*b>0) return(false);
  a=(x0-x2)*(y3-y2)-(y0-y2)*(x3-x2);
  b=(x1-x2)*(y3-y2)-(y1-y2)*(x3-x2);
  if(a*b>0) return(false);
  return(true);
}

/** AUXILIARY FUNCTION FOR FUNCTION bool shape::intersection_test(triangulation &T2) */
void intersection_test_aux(shape &S,triangulation &T2,bool &intersection){
  if(intersection==true) return;
  if(S.T.intersection_test(T2)==true){
     intersection=true;
     return;
  }
  for(int k=0;k<(int) S.shapeChild.size();k++){
   intersection_test_aux(S.shapeChild[k],T2,intersection);
  }
}

/** FUNCTION TO CHECK IF A TRIANGULATION INTERSECT A SHAPE */
bool shape::intersection_test(triangulation &T2){
  bool intersection=false;
  intersection_test_aux((*this),T2,intersection);
  return intersection;
}


/**********************************************************
test to check if a shape intersects another shape
**********************************************************/
bool shape::intersection_test(shape &S2)
{
//  S2.extrema_update();
//  extrema_update();
  if(extrema.size()==2 && S2.extrema.size()==2){
    if(extrema[0].x>S2.extrema[1].x ) return(false);
    if(extrema[0].y>S2.extrema[1].y ) return(false);
    if(S2.extrema[0].x>extrema[1].x ) return(false);
    if(S2.extrema[0].y>extrema[1].y ) return(false);
  }

  vector <shape*> SpV;
  S2.build_shape_pointer_vector(SpV);
  //printf("SpV.size()=%d\n",SpV.size());
  for(int k=0;k<(int) SpV.size();k++){
     bool intersection2=intersection_test((*SpV[k]).T);
     if(intersection2==true) return true;
  }

  return(false);
}

/** AUXILIARY FUNCTION FOR FUNCTION bool shape::inclusion_test(triangulation &T2) */
void inclusion_test_aux(shape &S,triangulation &T2,bool &inclusion){
  if(inclusion==true) return;
  if(S.T.inclusion_test(T2)==true){
     inclusion=true;
     return;
  }
  for(int k=0;k<(int) S.shapeChild.size();k++){
   inclusion_test_aux(S.shapeChild[k],T2,inclusion);
  }
}

/** FUNCTION TO CHECK IF A TRIANGULATION IS INCLUDED IN A SHAPE */
bool shape::inclusion_test(triangulation &T2){
  bool inclusion=false;
  //inclusion_test_aux(T2,inclusion);
  inclusion_test_aux((*this),T2,inclusion);
  return inclusion;
}


/**********************************************************
test to check if a shape is included in current shape
**********************************************************/
bool shape::inclusion_test(shape &S2)
{
  if(extrema.size()==2 && S2.extrema.size()==2){
    if(extrema[0].x>S2.extrema[0].x ) return false;
    if(extrema[0].y>S2.extrema[0].y ) return false;
    if(S2.extrema[1].x>extrema[1].x ) return false;
    if(S2.extrema[1].y>extrema[1].y ) return false;
  }

  vector <shape*> SpV;
  S2.build_shape_pointer_vector(SpV);
  //printf("SpV.size()=%d\n",SpV.size());
  for(int k=0;k<(int) SpV.size();k++){
     bool inclusion2=inclusion_test((*SpV[k]).T);
     if(inclusion2==false) return false;
  }

  return(true);
}


/** SET CHILD SHAPE COLOR FROM A PALETTE */
void shape::set_color(
imageBMP &palette /** palette image */,
int min_contrast /** min color contrast between a child shape and its father */){
    for(int k=0;k<(int) shapeChild.size();k++){
        color c2=palette[ami_randI(palette.size())];
        for(int l=0;l<100;l++){
          if(abs((int) c.R_-c2.R_)>=min_contrast || abs((int) c.G_-c2.G_)>=min_contrast || abs((int) c.B_-c2.B_)>=min_contrast) break;
          c2=palette[ami_randI(palette.size())];
        }
        shapeChild[k].c=c2;
        shapeChild[k].set_color(palette,min_contrast);
    }
}


/** SET SHAPE COLOR FROM A PALETTE WITH A MINIMUM CONTRAST*/
void shape::set_color(
imageBMP &palette /** palette image */,
color &c0,
int min_contrast /** min color contrast between a child shape and its father */){
  int i=ami_randI(palette.size());//-1)* ((double) rand()/RAND_MAX);
  color c2=palette[i];
  //printf("i=%d c=(%d,%d,%d)\n",i,c2.B_,c2.G_,c2.R_);
  for(int l=0;l<100;l++){
    if(abs((int) c0.R_-c2.R_)>=min_contrast || abs((int) c0.G_-c2.G_)>=min_contrast || abs((int) c0.B_-c2.B_)>=min_contrast) break;
    i=ami_randI(palette.size());//(palette.size()-1)* ((double) rand()/RAND_MAX);
    c2=palette[i];
  }
  //printf("i=%d c=(%d,%d,%d)\n",i,c2.B_,c2.G_,c2.R_);
  c=c2;
 for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].set_color(palette,c2,min_contrast);
  }
}

/** WE CHANGE POINT SHAPE COORDINATES USING AN AFFINITY */
void shape::operator()(affinity A){
    //for(int k=0;k<p.size();k++){ p[k]=A*p[k];}
    for(int k=0;k<(int) T.p.size();k++){ T.p[k]=A*T.p[k];}
    for(int k=0;k<(int) shapeChild.size();k++){shapeChild[k](A);}
    extrema_update();
}

/** WE CHANGE POINT SHAPE COORDINATES USING A TRANSLATION */
void shape::operator()(point2d d){
    //for(int k=0;k<p.size();k++){ p[k]=A*p[k];}
    for(int k=0;k<(int) T.p.size();k++){ T.p[k]=T.p[k]+d;}
    for(int k=0;k<(int) shapeChild.size();k++){shapeChild[k](d);}
    if(extrema.size()!=2) extrema_update();
    else{
      extrema[0]=extrema[0]+d;
      extrema[1]=extrema[1]+d;
    }

}

/** setting tree_level to all shape tree */
void shape::tree_level_set(int k){
    tree_level=k;
    for(int m=0;m<(int) shapeChild.size();m++)  shapeChild[m].tree_level_set(k+1);
  }

/** putting all shape tree as a sequence */
void shape::build_shape_pointer_vector(vector <shape*> &SpV){
    SpV.push_back(this);
    for(int k=0;k<(int) shapeChild.size();k++){
      shapeChild[k].build_shape_pointer_vector(SpV);
    }
}

/** count the number of triangulations of the shape */
void shape::count_triangulations(unsigned int &N){
  if(T.p.size()>0) N++;
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].count_triangulations(N);
  }
  for(int k=0;k<(int) shapeBorder.size();k++){
    shapeBorder[k].count_triangulations(N);
  }
}

void shape::set_transparency(float &transparency2,int tree_level2){ /** if tree_level2<0 we set the value in all shape tree */
  if(tree_level2<0 || tree_level2==tree_level) transparency=transparency2;
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].set_transparency(transparency2,tree_level2);
  }
}

void shape::set_light_spot_center(int tree_level2){ /** if tree_level2<0 we set the value in all shape tree */
  //printf("extrema.size()=%d\n",extrema.size()); system("pause");
  if( (tree_level2<0 || tree_level2==tree_level) && extrema.size()>1){
    light_spot_center.x=extrema[0].x + (extrema[1].x-extrema[0].x)*rand()/RAND_MAX;
    light_spot_center.y=extrema[0].y + (extrema[1].y-extrema[0].y)*rand()/RAND_MAX;
  }
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].set_light_spot_center(tree_level2);
  }
}

void shape::set_light_spot(float &light_spot2,int tree_level2){ /** if tree_level2<0 we set the value in all shape tree */
  if( (tree_level2<0 || tree_level2==tree_level) && extrema.size()>1) light_spot=light_spot2/(extrema[1]-extrema[0]).norm();
  for(unsigned int k=0;k<shapeChild.size();k++){
    shapeChild[k].set_light_spot(light_spot2,tree_level2);
  }
}

void shape::set_tesselation(int tesselation2, int tree_level2){ /** if tree_level2<0 we set the value in all shape tree */
  if(tree_level2<0 || tree_level2==tree_level){ tesselation_palette_index=tesselation2; }
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].set_tesselation(tesselation2,tree_level2);
  }
}

void shape::set_keepFatherInclusion(int keepFatherInclusion2, int tree_level2){ /** if tree_level2<0 we set the value in all shape tree */
  if(tree_level2<0 || tree_level2==tree_level){ keepFatherInclusion=keepFatherInclusion2; }
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].set_keepFatherInclusion(keepFatherInclusion2,tree_level2);
  }
}

void shape::symmetry(point2d c,int Nangles){

   T=T.symmetry(c,Nangles);
   for(int m=0;m<(int) shapeChild.size();m++)  shapeChild[m].symmetry(c,Nangles);
}

/** point coordinate normalization in the interval [-1,1]x[-1,1] */
shape shape::normalize(){
  extrema_update();
  if(extrema.size()!=2) return shape();
  float xmin=extrema[0].x, xmax=extrema[1].x, ymin=extrema[0].y, ymax=extrema[1].y;

  /* We compute an affinity to normalize*/
  float dx = xmax-xmin;
  float dy = ymax-ymin;
  //printf("dx=%lf dy=%lf\n",dx,dy);
  if(dx<1e-20 || dy<1e-20) return shape();
  float scale=dx>dy?2./dx:2./dy;
  //affinity A(scale,0.,0.,scale,-xmin*scale-1.,-ymin*scale-1.);
  affinity A(scale,0.,0.,scale,-(xmin+xmax)*0.5*scale,-(ymin+ymax)*0.5*scale);
  shape S=(*this)*A;
  S.extrema_update();
  return S;

}

void shape::tree_size(int &tree_number){
  tree_number+=shapeChild.size();
  for(int k=0;k<(int) shapeChild.size();k++){
    shapeChild[k].tree_size(tree_number);
  }
}

shape shape::extract_main_shape(){
  shape S;
  S.T=T;
  S.extrema=extrema;
  S.shapeBorder=shapeBorder;
  S.nodes=nodes;
  // OBJECT RENDER PROPERTIES
  S.c=c;
  S.transparency=transparency;
  S.light_spot=light_spot;
  S.light_spot_center=light_spot_center;
  S.tesselation_palette_index=tesselation_palette_index;
  S.keepFatherInclusion=keepFatherInclusion;
  return S;
}

/// MONDRIAN TYPE RECTANGLE SUBDIVISION
shape shape::mondrian(float &max_diameter,int MaxTries)
{
  printf("init mondrian\n");
  if(extrema.size()!=2) extrema_update();
  if(max_diameter<=0.) return shape();

  int Nx=(extrema[1].x-extrema[0].x)/max_diameter;
  int Ny=(extrema[1].y-extrema[0].y)/max_diameter;
  if(Nx==0 || Ny==0) return shape();
  float dx=(extrema[1].x-extrema[0].x)/Nx;
  float dy=(extrema[1].y-extrema[0].y)/Ny;

  shape S;

  int Ntries=0;
  if(Nx<2 || Ny<3) return shape();
  vector< vector<int> > flat(Nx,vector<int>(Ny,0));

  for(int x=0;x<Nx;x++){
    for(int y=0;y<Ny;y++){
      float xf=extrema[0].x+dx*x;
      float yf=extrema[0].y+dy*y;
      point2d p1(xf,yf);
      point2d p3(xf+dx,yf+dy);
      if(T.inclusion_test((p1+p3)*0.5)<0) flat[x][y]=1;
    }
  }

  while(Ntries++<MaxTries){
    int Dx=3,Dy=3;
    if(Ntries>10){
      Dx=rand()%4;
      Dy=rand()%4;
      while( ((Dx+Dy)<=3 && Ntries<MaxTries/2) || (((Dx+Dy)<1 && Ntries>=MaxTries/2)) ){
        Dx+=rand()%2;
        Dy+=rand()%2;
      }
    }
    if(Nx<=Dx || Ny<=Dy) continue;
    int x=rand()%(Nx-Dx);
    int y=rand()%(Ny-Dy);

    for(int sx=0;sx<=Dx;sx++){
      for(int sy=0;sy<=Dy;sy++){
        int x2=x+sx;
        int y2=y+sy;
        if(flat[x2][y2]==1){
          Dx=-1;
          break;
        }
      }
    }
    if(Dx==-1) continue;
    for(int sx=0;sx<=Dx;sx++){
      for(int sy=0;sy<=Dy;sy++){
        flat[x+sx][y+sy]=1;
      }
    }
    float xf=extrema[0].x+dx*x;
    float yf=extrema[0].y+dy*y;
    //printf("xf=%lf yf=%lf sx=%d sy=%d\n",xf,yf,sx,sy); system("pause");
    point2d p1(xf,yf);
    point2d p3(xf+(Dx+1)*dx,yf+(Dy+1)*dy);
    S.shapeChild.push_back(shape(triangulation(p1,p3)));

  }

  for(int x=0;x<Nx;x++){
    for(int y=0;y<Ny;y++){
      if(flat[x][y]==1) continue;
      float xf=extrema[0].x+dx*x;
      float yf=extrema[0].y+dy*y;
      point2d p1(xf,yf);
      point2d p3(xf+dx,yf+dy);
      S.shapeChild.push_back(shape(triangulation(p1,p3)));
    }
  }
  printf("end mondrian\n");

  return S;
}


