/**
 * \file shape.h
 * \brief class  to store shapes in a tree structure
 * \author Luis Alvarez \n \n
*/


#ifndef SHAPE_H
#define SHAPE_H
#include "color.h"
#include <iostream>
#include <vector>
#include <math.h>
#include "bmp.h"
#include "triangulation.h"

#define ami_PI 3.14159265358979323846

#define ami_rand ((double) rand()/RAND_MAX)

using namespace std;


/**
 * \class  shape
 * \brief class  shape
 * \author Luis Alvarez
 */
class shape{
  public :
  int tree_level; /** level in the tree structure */
  triangulation T; /** triangles defining the shape */
  vector<point2d> extrema; /** all shape tree bounding box :  extrema[0]=(xmin,ymin) extrema[1]=(xmax,ymax) */
  vector<shape> shapeChild; /** children shapes */
  vector<shape> shapeBorder; /** Shapes which define the border */
  vector<short> nodes; /// sequence of nodes in the tree to identify the collection node of shapes where the shape belongs

  /// OBJECT RENDER PROPERTIES
  color c;
  float transparency=0; /** transparency value of the object */
  float light_spot=-1.; /** Brightness reduction factor in object light spot effect */
  point2d light_spot_center; /** Center for the light spot effect */
  int  tesselation_palette_index=-1; /** if <0 no tesselation, otherwise means the index of the palette used */
  int keepFatherInclusion=0; /** =0 means that son object are not drawn outside the father objet=-1 the opposite */

  /// BASIC CONSTRUCTOR
  shape(){}; /** default constructor */
  /** constructor from a triangulation 1**/
  shape(triangulation T2){T=T2; c.random(); extrema_update(); tree_level=0;}
  /** constructor from a triangulation 2**/
  shape(int tree_level2,triangulation T2){tree_level=tree_level2; T=T2; c.random(); extrema_update();}
  /** constructor from a triangulation 3**/
  shape(int tree_level2,triangulation T2,color c2){tree_level=tree_level2; T=T2; c=c2; extrema_update();}
  /** constructor from a triangulation and random color assignement from a palette**/
  shape(int tree_level2,triangulation T2,imageBMP &palette){
    tree_level=tree_level2; T=T2; c=palette[ami_randI(palette.size())]; extrema_update();
  }
  shape mondrian(float &max_diameter,int MaxTries);
  /// CLEAR A SHAPE
  void clear(){T.clear(); extrema.clear(); shapeChild.clear(); shapeBorder.clear(); nodes.clear(); }

  /// BASIC SETTING
  void set_transparency(float &transparency2,int tree_level2);
  void set_light_spot_center(int tree_level2);
  void set_light_spot(float &light_spot2,int tree_level2);
  void set_tesselation(int tesselation2, int tree_level2);
  void set_keepFatherInclusion(int keepFatherInclusion2, int tree_level2);

  /// BASIC TREE OPERATIONS
  void erase_node(int k){ if(k>=(int) shapeChild.size()) return; shapeChild.erase(shapeChild.begin(),shapeChild.begin()+k);}
  void add_node(shape &S2){ shapeChild.push_back(S2); shapeChild[shapeChild.size()-1].tree_level=tree_level+1; }
  void tree_level_set(int k); /** setting tree_level to all shape tree */
  void build_shape_pointer_vector(vector <shape*> &SpV); /** putting all shape tree as a sequence */
  void tree_size(int &tree_number);
  shape extract_main_shape();
  void count_triangulations(unsigned int &N);

  /// BASIC GEOMETRIC OPERATIONS
  void operator()(affinity A); /** WE CHANGE POINT SHAPE COORDINATES USING AN AFFINITY */
  shape operator * (affinity A){ shape S2=(*this); S2(A); return S2;}
  void operator()(point2d d); /** WE CHANGE POINT SHAPE COORDINATES USING A TRANSLATION */
  shape operator + (point2d d){ shape S2=(*this); S2(d); return S2;}
  shape operator * (float s){ affinity A(s,0.,0.,s,0.,0.); shape S2=(*this); S2(A); return S2;}
  shape normalize();

  /// BASIC UTILITIES
  void set_color(imageBMP &palette,int min_contrast); /** set all shape tree color with a min_contrast between child and father*/
  void set_color(imageBMP &palette,color &c2,int min_contrast); /** set all shape tree color with a min_contrast between child and father*/
  void extrema_update(); /** updating the extrema defining the all shape tree bounding box */
  void print();
  bool intersection_test(triangulation &T2);
  bool intersection_test(shape &S2);
  bool inclusion_test(triangulation &T2);
  bool inclusion_test(shape &S2);

  /// build an Nangles symmetry triangulation from the original one
  void symmetry(point2d c,int Nangles);

};

/// PROCEDURES TO DRAW SHAPES IN A BMP IMAGE
void draw_shape(shape &S,imageBMP &I,float scaleI2Tr,vector < vector < vector<color> > > &TessPalette,int &palette_index,
                vector<imageBMP> &PaletteDataBase,triangulation &T,vector<unsigned char> &keepFatherInclusionV,bool periodicity=false,int MaxVideoFrames=0); // draw shape in an image
void draw_shape(shape &S,imageBMP &I,float scaleI2Tr,vector<imageBMP> &PaletteDataBase,int OrderToDrawShapes=0,bool periodicity=false,int VideoFrameRate=0); // draw shape in an image



#endif
