/**
 * \file shape_generation.cpp
 * \brief ****
 * \author Luis Alvarez \n \n
*/


#include <iostream>
#include <vector>
#include <math.h>
#include "shape_generation.h"



#define ami_PI 3.14159265358979323846
#define CERO 1e-12

vector<float> random_ordered_vector(int VectorSize,float min_extrema,float max_extrema){
 if(VectorSize<=0) return vector<float>();
 vector<float> v(VectorSize);
 v[0]=min_extrema+(max_extrema-min_extrema)*rand()/RAND_MAX;
 float min_distance=(max_extrema-min_extrema)/VectorSize/4.;
 for(int k=1;k<(int) v.size();k++){
   float distance=0.;
   while(distance<min_distance){
     distance=min_distance+1;
     v[k]=min_extrema+(max_extrema-min_extrema)*rand()/RAND_MAX;
     for(int n=0;n<k;n++){
       float temp=fabs(v[k]-v[n]);
       if(temp<distance) distance=temp;
     }
   }
 }
 // WE ORDER THE VECTOR
 for(int k=0;k<(int) v.size();k++){
    for(int n=k+1;n<(int) v.size();n++){
      if(v[k]<v[n]){
        float temp=v[k]; v[k]=v[n]; v[n]=temp;
      }
    }
 }
 return v;
}
/**********************************************************
SHAPE GENERATOR
**********************************************************/
shape shape_generator::new_shape(int &shape_database,int shape_index,float zoom){
  triangulation T;
  if(shape_index<0){
    shape_index=rand();
  }
  shape_database=shape_database%7;
  switch(shape_database) {
    case 0 : /// basic shapes
      switch(shape_index%7){
        case 0 : /// triangle
          //T=triangulation();
          //printf("case 0-0-0 : triangle\n");
          T.p.push_back(point2d(1.,0.));
          T.p.push_back(point2d(cos(2.*ami_PI/3.),sin(2.*ami_PI/3.)));
          T.p.push_back(point2d(cos(4.*ami_PI/3.),sin(4.*ami_PI/3.)));
          T.fr.push_back(1); T.fr.push_back(1); T.fr.push_back(1);
          T.t.push_back(triangle(0,1,2));
          T.n.push_back(triangle(-1,-1,-1));
          return shape(T);
        case 1 : /// square
          //printf("case 0-0-1 : square\n");
          return shape(triangulation(point2d(-1.,-1.),point2d(1.,1.)));
        case 2 : /// circle
          //printf("case 0-0-2 : circle\n");
          return shape(triangulation(point2d(0.,0.),1.,60));
        case 3 : /// hexagon
          //printf("case 0-0-3 : hexagon\n");
          return shape(triangulation(point2d(0.,0.),1.,6));
        case 4 : /// star
          //printf("case 0-0-4 : star\n");
          {
            vector<point2d> pV;
            for(int k=0;k<10;k++){
            float a=2.*k*ami_PI/10.;
            if(k%2==0) pV.push_back(point2d(cos(-a),sin(-a)));
            else pV.push_back(point2d(cos(-a)*0.4,sin(-a)*0.4));
            //printf("p[%d]=(%lf,%lf) norm=%lf\n",k,pV[k].x,pV[k].y,pV[k].norm());
            }
            //system("pause");
            return shape(triangulation(pV));
          }
        case 5 : /// polygon of Npoints
          //printf("case 0-0-5 : polygon\n");
          {
            int N=4+7*rand()/RAND_MAX;
            vector<point2d> pV(N);
            vector<float> a=random_ordered_vector(N,0.,2.*ami_PI);
            for(int k=0;k<N;k++){
              pV[k]=(point2d(cos(a[k]),sin(a[k])));
            }
            return shape(triangulation(pV));
          }
        case 6 : /// rectangle band with a random strip thick
          //printf("case 0-0-6 : rectangle band\n");
          {
            float strip_thick=0.3+0.4*rand()/RAND_MAX;
            return shape(triangulation(strip_thick,point2d(-1.,-1.),point2d(1.,1.)));
          }

        case 7 :
          break;
        case 8 :
          break;
        case 9 :
          break;
        case 10 :
          break;
        case 11 :
          break;
      }
      break;
    case 1 : /// constructor based on segment random evolution
      switch(shape_index%3){
        case 0 :
          {
            //printf("case 0-1-0 : brownian caotic evolution with variable segment thick \n");
            float radius=1.+1.*rand()/RAND_MAX;
            triangulation T0(point2d(-radius,-radius),point2d(radius,radius));
            float r=0.1+0.2*rand()/RAND_MAX;
            float a=-0.01+0.02*rand()/RAND_MAX;
            float da=0.5+1.*rand()/RAND_MAX;
            float dt=0.1,dr=0.2*rand()/RAND_MAX;
            triangulation T1(T0,point2d(0.,0.),dt,a,-da+a,da+a,r,-dr,dr,200);
            T1.put_horizontal();
            T1=T1-T1.p[T1.p.size()/2];
            T1.rebuild(dt);
            T1.rebuildOnlyContour();
            return shape(T1);
          }
        case 1 :
          {
            //printf("case 0-1-1 : brownian smooth evolution with fix segment thick\n");
            float radius=0.5+3.5*rand()/RAND_MAX;
            triangulation T0(point2d(-radius,-radius),point2d(radius,radius));
            float r=0.025,da=0.2,a=0.,dt=0.1,dr=0.;
            triangulation T1(T0,point2d(0.,0.),dt,a,-da+a,da+a,r,-dr,dr,5000);
            T1.put_horizontal();
            T1=T1-T1.p[T1.p.size()/2];
            T1.rebuild();
            return shape(T1);
          }
        case 2 :
          {
            //printf("case 0-1-2: brownian caotic evolution with fix segment thick\n");
            float radius=0.5+3.5*rand()/RAND_MAX;
            triangulation T0(point2d(-radius,-radius),point2d(radius,radius));
            float r=0.025,da=1.,a=0.,dt=0.1,dr=0.;
            triangulation T1(T0,point2d(0.,0.),dt,a,-da+a,da+a,r,-dr,dr,200);
            T1.put_horizontal();
            T1=T1-T1.p[T1.p.size()/2];
            T1.rebuild();
            return shape(T1);
          }

      }
      break;
    case 2 : /// Kandisky circle
      //printf("case 0-2 : Kandisky circle\n");
      Kc=kandisky_circle(shape_index%1);
      return (Kc.new_shape()).normalize();
    case 3 : /// Kandisky rectangle
      //printf("case 0-3 : Kandisky rectangle\n");
      Kr=kandisky_rectangle(shape_index%1);
      return (Kr.new_shape()).normalize();
    case 4 : /// tree
      //printf("case 0-4 : Kandisky tree \n");
      {
        shape S=shape_generator().new_shape(301);
        float zoom=3.+3.*rand()/RAND_MAX;
        contouring(S,S,-0.1,100,10,0.001,zoom);
        zoom=0.75+0.5*rand()/RAND_MAX;
        contouring(S,S,0.4,300,30,0.01,zoom);
//        S.print();
//        printf("S.shapeChild.size()=%d\n",S.shapeChild.size());
//        system("pause");
        for(int k=0;k<(int) S.shapeChild.size();k++){
          for(int n=0;n<(int) S.shapeChild[k].T.p.size();n++){
            if(S.shapeChild[k].T.p[n].y-0.1<S.T.p[0].y){
              S.shapeChild.erase(S.shapeChild.begin()+k);
              k--;
              break;
            }
          }
        }
        return (S.normalize());
      }
    case 5 :
    {
      //printf("case 0-5 : rosette  \n");
      Kr=kandisky_rectangle(shape_index%1);
      shape S2=Kr.new_shape();
      S2.symmetry(point2d(0.,0.),16);
      return S2.normalize();
    }
    case 6 :
    {
      shape S2=NodeTreeGeneration(shape_index%10,zoom);
      //S2.T.rebuildOnlyContour();
      return S2.normalize();
    }
    case 7 :
      switch(shape_index%6){
        case 0 :
        case 1 :
          break;
      }
      break;
    case 8 :
      switch(shape_index%6){
        case 0 :
        case 1 :
           break;
      }
      break;
    case 9 :
      switch(shape_index%6){
        case 0 :
        case 1 :
            break;
      }
      break;
  }
  return shape();;
};

shape NodeTreeGeneration(int type,float zoom){

  switch(type){

    case 0 : /// water drop
    {
      float dmin=64.;
      triangulation T0=triangulation(point2d(0.,0.),point2d(1024.,1024.));

      float R0=150+100*myrand(1.);

      NodeTree nT;

      nT.n_.push_back(point2d(512.,dmin+R0));
      nT.i_.resize(1);
      nT.r_.push_back(R0);
      nT.s_.push_back(1.);

      nT.insert_point( point2d(512.+(-1.+2.*myrand(1.))*30,1024.-dmin) , 1., 0.1, 0);

      vector<double> angles=nT.angle_range(13);
      int posa=myrand(1.)*angles.size();
      posa=posa%angles.size();
      nT.insert_point( nT.n_[1]-point2d(cos(angles[posa]),sin(angles[posa]))*3 , 5., 1., 0,1);

      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return(shape(nT.triangulation_generation(10.)));

    }

    case 1 : /// spyder
    {
       float dmin=64.;
       /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
       float R0=512-dmin;
       //float T0y[5]={348.,212.,0.,-238.,-448.};
       float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
       float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
       triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

       NodeTree nT;
       nT.n_.push_back(point2d(512.,512.));
       nT.i_.resize(1);
       nT.r_.push_back(70);
       nT.s_.push_back(1);

       float min_segment_length=200.,max_segment_length=250.;
       float distance_segment_factor=2.;
       float min_radius=30.,max_radius=50.;
       float distance_node_factor=4.;
       float smooth_factor=0.5;
       int type=0;
       float gamma=1.;
       int Nbranches=1;
       vector<double> angles=nT.angle_range(100);
       unsigned int MaxNnodes=100000;

       nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
      );

      min_radius=3.;max_radius=3;
      distance_segment_factor=20.;
      min_segment_length=200.;
      max_segment_length=250.;
      distance_node_factor=1500.;
      Nbranches=1;
      unsigned int MaxNeigbors=2;
      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes,
        MaxNeigbors
      );
      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return(shape(nT.triangulation_generation(10.)));

    }

    case 2 : /// matisse tree
    {

      NodeTree nT;
      zoom=0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;
      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0R[5]={10.,128.,300.,200.,150};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

      float min_segment_length=150.*zoom,max_segment_length=200.*zoom;
      float distance_segment_factor=10.;
      float min_radius=50.*zoom,max_radius=50*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=1.5;
      int Nbranches=(2.+myrand(1)*3)/zoom;
      vector<double> angles=nT.angle_range(3);
      //for(unsigned int k=0;k<angles.size();k++) angles[k]=M_PI/2+0.5*(angles[k]-M_PI/2);
      unsigned int MaxNnodes=100000;


      nT.n_.push_back(point2d(512.,dmin+30.));
      nT.i_.resize(1);
      nT.r_.push_back(min_radius);
      nT.s_.push_back(smooth_factor);

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );


      gamma=1.5;
      Nbranches=1;
      angles=nT.angle_range(19);
      distance_segment_factor=2.5;
      min_radius=30.*zoom; max_radius=40.*zoom;
      min_segment_length=100.*zoom;
      max_segment_length=150.*zoom;
      smooth_factor=0.75;
      int MaxNeigbors=3;


      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes,
        MaxNeigbors
     );

      Nbranches=100;
      angles=nT.angle_range(19);
      distance_segment_factor=2.;
      min_radius=40*zoom; max_radius=60.*zoom;
      distance_node_factor=3.;
      min_segment_length=100.*zoom;
      max_segment_length=150.*zoom;


      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );

     if(myrand(1.)<0.5){
       for(unsigned int k=0;k<nT.n_.size();k++){
          if(nT.i_[k].size()==1) {
            nT.r_[k]=1;
            nT.s_[k]=0.1;
            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
          }
        }
     }

      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        //system("pause");
        return shape();
      }

      return shape(nT.triangulation_generation(10.));
     }

    case 3 : /// ramifications from a center
    {

      NodeTree nT;
      zoom=0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;

      bool round_termination = myrand(1.)<0.5?true:false;
      float join_probality=myrand(1.);
      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

      nT.clear();
      nT.n_.push_back(point2d(512.+(2.*myrand(1.)-1)*R0*0.5,512.+(2.*myrand(1.)-1)*R0*0.5));
      nT.i_.resize(1);
      nT.r_.push_back(50*1.5*zoom);
      nT.s_.push_back(0.75);

      float min_segment_length=200.*zoom,max_segment_length=250.*zoom;
      float distance_segment_factor=2.5;
      float min_radius=50.*zoom,max_radius=60*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=1.;
      int Nbranches=100;
      vector<double> angles=nT.angle_range(100);
      unsigned int MaxNnodes=100000;

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );

     nT.joint_extrema(join_probality);

     if(round_termination==false){
        for(unsigned int k=0;k<nT.n_.size();k++){
          if(nT.i_[k].size()==1) {
            nT.r_[k]=1;
            nT.s_[k]=0.1;
            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
          }
        }
      }


      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return shape(nT.triangulation_generation(10.));
     }

    case 4 : /// matisee seawood
    {

      NodeTree nT;
      zoom=2.*0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;

      bool round_termination = myrand(1.)<0.5?true:false;
      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

      nT.clear();
      nT.n_.push_back(point2d(512.,512.-0.9*R0));
      nT.i_.resize(1);
      nT.r_.push_back(50*zoom);
      nT.s_.push_back(0.75);

      int il,ir;

      {
        for(unsigned int k=1;(float) k<4./zoom;k++){
          nT.insert_point(nT.n_[nT.n_.size()-1]+point2d(90.*zoom,myrand(1.)*20*zoom),50.*zoom,0.75,nT.n_.size()-1);
          if(nT.n_[nT.n_.size()-1].x-1.2*T0R[4]>512.) break;
        }
        ir=nT.n_.size()-1;
        nT.insert_point(nT.n_[0]+point2d(-90.*zoom,myrand(1.)*20*zoom),50.*zoom,0.75,0);
        for(unsigned int k=2;(float) k<4./zoom;k++){
          nT.insert_point(nT.n_[nT.n_.size()-1]+point2d(-90.*zoom,myrand(1.)*20*zoom),50.*zoom,0.75,nT.n_.size()-1);
          if(nT.n_[nT.n_.size()-1].x+T0R[4]*1.2<512.) break;
        }
        il=nT.n_.size()-1;
      }
//      nT.print();
//      system("pause");

      float min_segment_length=100.*zoom,max_segment_length=150.*zoom;
      float distance_segment_factor=2.;
      float min_radius=30.*zoom,max_radius=40*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=1.5;
      int Nbranches=6/zoom;
      vector<double> angles=nT.angle_range(11);
      unsigned int MaxNnodes=100000;

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );

     min_radius=40.*zoom;max_radius=50*zoom;
     distance_segment_factor=2.;
     distance_node_factor=4.;
     Nbranches=100;
     nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );

     if(round_termination){
       if(nT.i_[ir].size()>1) nT.insert_point(nT.n_[ir]+point2d(100.*zoom,0.),40.*zoom,0.75,ir);
       if(nT.i_[il].size()>1) nT.insert_point(nT.n_[il]-point2d(100.*zoom,0.),40.*zoom,0.75,il);
     }
     else{
       if(nT.i_[ir].size()>1) nT.insert_point(nT.n_[ir]+point2d(200.*zoom,0.),1,0.75,ir);
       if(nT.i_[il].size()>1) nT.insert_point(nT.n_[il]-point2d(200.*zoom,0.),1,0.75,il);
     }


      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      if(round_termination==false){
        for(unsigned int k=0;k<nT.n_.size();k++){
          if(nT.i_[k].size()==1) {
            nT.r_[k]=1;
            nT.s_[k]=0.1;
            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
          }
        }
      }


      return shape(nT.triangulation_generation(10.));
     }

    case 5 : /// matisee seawood bis
    {

      NodeTree nT;
      zoom=2.*0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;

      bool round_termination = myrand(1.)<0.5?true:false;
      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

            nT.clear();
      nT.n_.push_back(point2d(512.,512.-0.9*R0));
      nT.i_.resize(1);
      nT.r_.push_back(50*zoom);
      nT.s_.push_back(0.75);

      float min_segment_length=100.*zoom,max_segment_length=150.*zoom;
      float distance_segment_factor=2.;
      float min_radius=30.*zoom,max_radius=40*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=2.;
      int Nbranches=1000;
      vector<double> angles=nT.angle_range(7);
      unsigned int MaxNnodes=100000;
      unsigned int MaxNeigbors=2;

      vector<int> posR(1,0);
      vector<int> posL(1,0);
      int direction=1.;
      while(true){
        //printf("pos=%d\n",pos); system("pause");

        nT.insert_nodes(
          T0,
          min_segment_length,max_segment_length,
          distance_segment_factor,
          min_radius,max_radius,
          distance_node_factor,
          smooth_factor,
          type,
          gamma,
          Nbranches,
          angles,
          MaxNnodes,
          MaxNeigbors
        );

        int pos2=direction==1?posR[posR.size()-1]:posL[posL.size()-1];
        point2d  p;
        if(pos2==0) p=nT.n_[pos2]+point2d(90.*zoom*direction,(0.1+0.6*myrand(1.))*35.*zoom);
        else p=nT.n_[pos2]+point2d(90.*zoom*direction,(0.2+1.2*myrand(1.))*35.*zoom);
        if(T0.inclusion_test(p)<0){
          if(direction==-1) break;
          direction=-1;
          p=nT.n_[0]+point2d(90.*zoom*direction,(0.1+0.6*myrand(1.))*35.*zoom);
        }
        if(direction==1){
          nT.insert_point(p,50.*zoom,0.75,posR[posR.size()-1]);
          posR.push_back(nT.n_.size()-1);
        }
        else{
          nT.insert_point(p,50.*zoom,0.75,posL[posL.size()-1]);
          posL.push_back(nT.n_.size()-1);
        }

      }

      MaxNeigbors=3;
      angles=nT.angle_range(11);
      distance_segment_factor=2.;
      distance_node_factor=7.;
      min_segment_length=150.*zoom;
      max_segment_length=200.*zoom;

      nT.insert_nodes(
          T0,
          min_segment_length,max_segment_length,
          distance_segment_factor,
          min_radius,max_radius,
          distance_node_factor,
          smooth_factor,
          type,
          gamma,
          Nbranches,
          angles,
          MaxNnodes,
          MaxNeigbors
      );

      for(unsigned int k=posL.size()-1;k>0;k--){
        if(nT.i_[posL[k]].size()==1) nT.erase_point(posL[k]);
      }
      for(unsigned int k=posR.size()-1;k>0;k--){
        if(nT.i_[posR[k]].size()==1) nT.erase_point(posR[k]);
      }

      if(round_termination==true){
        for(unsigned int k=0;k<nT.n_.size();k++){
          if(nT.i_[k].size()==1) {
            nT.r_[k]=1;
            nT.s_[k]=0.1;
            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
          }
       }
      }

      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return shape(nT.triangulation_generation(10.));
     }

    case 6 : /// labyrinth
    {

      NodeTree nT;
      zoom=0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;

     float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);

      nT.clear();
      nT.n_.push_back(point2d(512.+(2.*myrand(1.)-1)*R0*0.5,512.+(2.*myrand(1.)-1)*R0*0.5));
      nT.i_.resize(1);
      nT.r_.push_back(50*zoom);
      nT.s_.push_back(0.);

      float min_segment_length=200.*zoom,max_segment_length=200.*zoom;
      float distance_segment_factor=2.5;
      float min_radius=50.*zoom,max_radius=50*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.;
      int type=0;
      float gamma=1.;
      int Nbranches=100;
      unsigned int MaxNnodes=100000;
      vector<double> angles; //=nT.angle_range(100);
      angles.push_back(0.);
      angles.push_back(M_PI/2);
      angles.push_back(M_PI);
      angles.push_back(3*M_PI/2);
      if(myrand(1.)<0.5){
        angles.push_back(M_PI/4);
        angles.push_back(3*M_PI/4);
        angles.push_back(5*M_PI/4);
        angles.push_back(7*M_PI/4);
      }

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes
     );

      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return shape(nT.triangulation_generation(10.));
     }

    case 7 : /// random connections
    {

      NodeTree nT;
      zoom=0.8*0.0600749531835024/zoom;
      if(zoom>1.5) zoom=1.5;
      else if(zoom<0.2) zoom=0.2;

      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);


      float min_segment_length=200.*zoom,max_segment_length=250.*zoom;
      float distance_segment_factor=2.;
      float min_radius=20.*zoom,max_radius=100*zoom;
      float distance_node_factor=4.;
      float max_distance_node_factor=32.;
      float min_smooth_factor=0.1,max_smooth_factor=1;
      float gamma=1.;
      unsigned int MaxNnodes=1000;
      double min_angle_distance=M_PI/16.;
      unsigned int MaxNeigbors=1000;
      unsigned int MaxAttempts=1000;
      nT.random_generation(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        max_distance_node_factor,
        min_smooth_factor,max_smooth_factor,
        gamma,
        min_angle_distance,
        MaxNnodes,
        MaxNeigbors,
        MaxAttempts
      );

      if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

      return shape(nT.triangulation_generation(10.));
     }
     case 8 : /// symmetry
    {

      int SymmetryType=1+rand()%3;
      zoom=0.8*0.0600749531835024/zoom;
      if(zoom>1) zoom=1.;
      else if(zoom<0.2) zoom=0.2;
      float dmin=64.;
      /// BUILD A DECAGON TO CONSTRAINT THE SHAPE
      float R0=512-dmin;
      //float T0y[5]={348.,212.,0.,-238.,-448.};
      float T0y[5]={R0*sinf(2*M_PI/5),R0*sinf(M_PI/5),0,-R0*sinf(M_PI/5),-R0*sinf(2*M_PI/5)};
      float T0R[5]={R0*cosf(2*M_PI/5),R0*cosf(M_PI/5),R0,R0*cosf(M_PI/5),R0*cosf(2*M_PI/5)};
      triangulation T0=triangulation(point2d(512.,512.),T0y,T0R);



      NodeTree nT;

      float min_segment_length=200.*zoom,max_segment_length=250.*zoom;
      while(true){
        point2d p(512.+(2.*myrand(1.)-1)*R0*0.5,512.+(2.*myrand(1.)-1)*R0*0.5);
        if(nT.InsideSymmetricRegion(p.x,p.y,512.,512.,SymmetryType)==false) continue;
        nT.point_line_projection(p,point2d(512.,512.),min_segment_length*0.5,SymmetryType);
        nT.n_.push_back(p);
        nT.i_.resize(1);
        nT.r_.push_back(50*1.5*zoom);
        nT.s_.push_back(0.75);
        break;
      }


      float distance_segment_factor=2.5;
      float min_radius=50.*zoom,max_radius=60*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=1.;
      int Nbranches=1000;
      vector<double> angles=nT.angle_range(100);
      unsigned int MaxNnodes=10000;
      unsigned int MaxNeigbors=100;

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes,
        MaxNeigbors,
        SymmetryType
     );

      //nT.print();
      if(myrand(1.)<0.5){
        for(unsigned int k=0;k<nT.n_.size();k++){
          if(nT.i_[k].size()==1) {
            nT.r_[k]=1;
            nT.s_[k]=0.1;
            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
          }
       }
      }

      return shape(nT.triangulation_generation(10.));

    }
    case 9 : /// periodicity
    {

      NodeTree nT;
      zoom=0.5*0.8*0.0600749531835024/zoom;
      if(zoom>1.) zoom=1.;
      else if(zoom<0.2) zoom=0.2;

      int SymmetryType=4+rand()%3;
      float dmin=128.;

      triangulation T0(point2d(512.-dmin,512.-dmin), point2d(512.+dmin,512.+dmin));



      nT.clear();

      bool labyrinth=myrand(1.)<0.?true:false;
      labyrinth=false;
      float min_segment_length=200.*zoom,max_segment_length=250.*zoom;
      if(labyrinth==true) max_segment_length=min_segment_length;
      while(true){
        point2d p(512.+dmin*myrand(1.),512.+dmin*myrand(1.));
        if(labyrinth==true) p.y=p.x;
        if(nT.InsideSymmetricRegion(p.x,p.y,512.,512.,SymmetryType)==false) continue;
        nT.point_line_projection(p,point2d(512.,512.),min_segment_length*0.5,SymmetryType);
        nT.n_.push_back(p);
        nT.i_.resize(1);
        nT.r_.push_back(50*zoom);
        if(labyrinth==true) nT.s_.push_back(0.);
        else nT.s_.push_back(0.75);
        break;
      }


      float distance_segment_factor=2.5;
      float min_radius=50.*zoom,max_radius=50*zoom;
      float distance_node_factor=5.;
      float smooth_factor=0.75;
      int type=0;
      float gamma=1.;
      int Nbranches=1000;
      vector<double> angles=nT.angle_range(100);
      unsigned int MaxNnodes=10000;
      unsigned int MaxNeigbors=100;

      if(labyrinth==true){
        smooth_factor=0.;
        angles.clear();
        angles.push_back(0.);
        angles.push_back(M_PI/2);
        angles.push_back(M_PI);
        angles.push_back(3*M_PI/2);
        if(myrand(1.)<0.5){
          angles.push_back(M_PI/4);
          angles.push_back(3*M_PI/4);
          angles.push_back(5*M_PI/4);
          angles.push_back(7*M_PI/4);
        }
      }

      nT.insert_nodes(
        T0,
        min_segment_length,max_segment_length,
        distance_segment_factor,
        min_radius,max_radius,
        distance_node_factor,
        smooth_factor,
        type,
        gamma,
        Nbranches,
        angles,
        MaxNnodes,
        MaxNeigbors,
        SymmetryType
     );

     float length=256.;
     float dx[8]={1.,0.,1.,-1., 0., 1.,-1.,-1.};
     float dy[8]={0.,1.,1., 0.,-1.,-1., 1.,-1.};

     NodeTree nT0=nT;
     for(int m=0;m<1;m++){
        nT.add(nT0.translation(point2d(dx[m],dy[m])*length));
     }
     if(nT.checking()==false){
        printf("nT.checking()=%d\n",false);
        return shape();
      }

//      //nT.print();
//      bool arrow_termination=false;
//      if(myrand(1.)<0.5 && labyrinth==false){
//        arrow_termination=true;
//        for(unsigned int k=0;k<nT.n_.size();k++){
//          if(nT.i_[k].size()==1) {
//            nT.r_[k]=1;
//            nT.s_[k]=0.1;
//            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
//          }
//       }
//      }
//      if(arrow_termination==true){
//        for(unsigned int k=0;k<nT.n_.size();k++){
//          if(nT.i_[k].size()==1) {
//            nT.r_[k]=1;
//            nT.s_[k]=0.1;
//            nT.n_[k]=nT.n_[nT.i_[k][0]]+(nT.n_[k]-nT.n_[nT.i_[k][0]])*0.75;
//          }
//       }
//      }

      triangulation T;
      if(labyrinth==false) T=nT.triangulation_generation(10.);
      else T=nT.triangulation_generation(10.,false);
      printf("N.nodos 1=%d,  ",(int) nT.n_.size() );

      return shape(nT.triangulation_generation(10.));
     }

   }
   return shape();
}


shape shape_generator::new_shape(int shape_type_new){

  /// TRIANGLE
  if(shape_type_new==0){
    triangulation T;
    T.p.push_back(point2d(1.,0.));
    T.p.push_back(point2d(cos(2.*ami_PI/3.),sin(2.*ami_PI/3.)));
    T.p.push_back(point2d(cos(4.*ami_PI/3.),sin(4.*ami_PI/3.)));
    T.fr.push_back(1); T.fr.push_back(1); T.fr.push_back(1);
    T.t.push_back(triangle(0,1,2));
    T.n.push_back(triangle(-1,-1,-1));
    return shape(T);
  }
  /// SQUARE
  else if(shape_type_new==1){
    return shape(triangulation(point2d(-1.,-1.),point2d(1.,1.)));
  }
  /// CIRCLE
  else if(shape_type_new==2){
    return shape(triangulation(point2d(0.,0.),1.,20));
  }

 /// KANDISKY CIRCLE : shape_type_new in [100,200)
 if(99<shape_type_new && shape_type_new<200){
    shape_type_new-=100;
    if(shape_type_new==0) return Kc.new_shape();
    Kc=kandisky_circle(shape_type_new-1);
    return Kc.new_shape();
 }

 /// KANDISKY RECTANGLE shape_type_new in [200,300)
 else if (199<shape_type_new && shape_type_new<300){
    shape_type_new-=200;
    if(shape_type_new==0) return Kr.new_shape();
    Kr=kandisky_rectangle(shape_type_new-1);
    return Kr.new_shape();
 }
 /// LEAF shape_type_new in [300,400)
  else if (299<shape_type_new && shape_type_new<400){
    shape_type_new-=300;
    if(shape_type_new==0) return Leaf.new_shape();
    Leaf=leaf(shape_type_new-1);
    return Leaf.new_shape();
 }

 return shape();

}

/// OBSOLETO
shape shape_generator::new_shape(int shape_type_new,triangulation &T1,int index_triangle,int index_segment,float min_distance,float zoom){

 shape S;
 /// KANDISKY CIRCLE : shape_type_new in [100,200)
 if(99<shape_type_new && shape_type_new<200){
    shape_type_new-=100;
    //if(shape_type_new==0) return Kc.new_shape();
    Kc=kandisky_circle(shape_type_new-1);
    S=Kc.new_shape();
    //printf("S.T.p.size()=%d\n",S.T.p.size()); system("pause");
 }

 /// KANDISKY RECTANGLE shape_type_new in [200,300)
 else if (199<shape_type_new && shape_type_new<300){
    shape_type_new-=200;
    //if(shape_type_new==0) return Kr.new_shape();
    Kr=kandisky_rectangle(shape_type_new-1);
    S=Kr.new_shape();
 }
 /// LEAF shape_type_new in [300,400)
  else if (299<shape_type_new && shape_type_new<400){
    shape_type_new-=300;
    //if(shape_type_new==0) return Leaf.new_shape();
    Leaf=leaf(shape_type_new,T1,index_triangle,index_segment);
    //printf("Leaf.dt=%lf Leaf.Niter=%d\n",Leaf.dt,Leaf.Niter); system("pause");
    S=Leaf.new_shape();
 }

 if(S.T.p.size()==0) return shape();
 if(zoom!=1.) S=S*zoom;
 affinity A=connectT12T0(T1,index_triangle,index_segment,S.T,1,2,min_distance,1);
 return S*A;

}



//// GENERATOR OF KANDISKY CIRCLE
shape kandisky_circle_generator(
float radius /** radius of kandisky circle */,
int Nlevels /** number of inside circles */,
int CircleNpoints /** number of points to discretize the circle*/)
{
  point2d center(0.,0.);
  float angle0=2.*M_PI*((float) rand()/RAND_MAX-0.5);
  float par0=0.2*((float) rand()/RAND_MAX);
  float angle1=2.*M_PI*((float) rand()/RAND_MAX-0.5);
  float par1=0.2+0.2*((float) rand()/RAND_MAX);
  triangulation T(center,radius,CircleNpoints);
  //printf("T.p.size()=%d CircleNpoints=%d \n",T.p.size(),CircleNpoints);

  shape S(1,T.deformation(center,angle0,par0,angle1,par1));
  shape *Sp;
  Sp=&S;
  for(int l=1;l<Nlevels;l++){
    triangulation T2(center,radius-radius*l/(Nlevels+1.),CircleNpoints);
    bool inclusion=false;
    while(inclusion==false){
      angle0=2.*M_PI*((float) rand()/RAND_MAX-0.5);
      par0=0.2*((float) rand()/RAND_MAX);
      angle1=2.*M_PI*((float) rand()/RAND_MAX-0.5);
      par1=0.2+0.2*((float) rand()/RAND_MAX);
      triangulation T3=T2.deformation(center,angle0,par0,angle1,par1); //deformed circle
      //triangulation T3=T2;
      inclusion=(*Sp).inclusion_test(T3); //we check if each circle is included in its father circle
      if(inclusion==true){
        (*Sp).shapeChild.push_back(shape((*Sp).tree_level+1,T3)); // we add the new circle as a child shape
        Sp=&((*Sp).shapeChild[0]); // we update the shape pointer
      }
    }
  }
  //S.extrema_update();

  return S;
}

// GENERATOR OF KANDISKY RECTANGLE
shape kandisky_rectangle_generator(
float width /** rectangle width */,
float height /** rectangle height */,
int Nlevels /** number of inside rectangles */,
float min_distance /** min distance between points of the rectangle */)
{
  point2d center(0.,0.);
  float angle0=2.*M_PI*((float) rand()/RAND_MAX-0.5);
  float par0=0.2*((float) rand()/RAND_MAX);
  float angle1=2.*M_PI*((float) rand()/RAND_MAX-0.5);
  float par1=0.2+0.2*((float) rand()/RAND_MAX);
  //triangulation T(center,radius,CircleNpoints);
  triangulation T(point2d(-width/2,-height/2),point2d(width/2,height/2),min_distance);

  shape S(1,T.deformation2(center,angle0,par0,angle1,par1));
  shape *Sp;
  Sp=&S;
  for(int l=1;l<Nlevels;l++){
    //triangulation T2(center,radius-radius*l/(Nlevels+1.),CircleNpoints);
    float width2=width-width*l/(Nlevels+1.);
    float height2=height-height*l/(Nlevels+1.);
    triangulation T2(point2d(-width2/2,-height2/2),point2d(width2/2,height2/2),min_distance);
    bool inclusion=false;
    while(inclusion==false){
      angle0=2.*M_PI*((float) rand()/RAND_MAX-0.5);
      par0=0.2*((float) rand()/RAND_MAX);
      angle1=2.*M_PI*((float) rand()/RAND_MAX-0.5);
      par1=0.2+0.2*((float) rand()/RAND_MAX);
      triangulation T3=T2.deformation2(center,angle0,par0,angle1,par1); //deformed circle
      inclusion=(*Sp).inclusion_test(T3); //we check if each circle is included in its father circle
      if(inclusion==true){
        (*Sp).shapeChild.push_back(shape((*Sp).tree_level+1,T3)); // we add the new circle as a child shape
        Sp=&((*Sp).shapeChild[0]); // we update the shape pointer
      }
    }
  }
  return S;
}


// GENERATION OF A KANDISKY RECTANGLE FRAME
shape kandisky_rectangle_frame_generator(
point2d  pmin /** frame window min point coordinates */,
point2d pmax /** frame window max point coordinates */,
int NCircles_x /** number of  horizontal kandisky rectangles taking into account the corners */,
int NCircles_y /** number of vertical kandisky rectangles without taking into account the corners */,
int Nlevels /** number of levels (inside rectangles) of the kandisky rectangle */,
float min_distance /** min distance between rectangle points */)
{
  if(NCircles_x==0 && NCircles_y==0) return shape();
  float hx,hy;
  if(NCircles_x!=0) hx=(pmax.x-pmin.x)/NCircles_x;
  else hx=pmax.x-pmin.x;
  if(NCircles_y!=0) hy=(pmax.y-pmin.y-2*hx)/NCircles_y;
  else hy=pmax.y-pmin.y;
  shape S;
  triangulation T0(pmin,pmax);
  triangulation T1(pmin+point2d(hx,hy),pmax-point2d(hx,hy));
  if(NCircles_x!=0 && NCircles_y!=0) S=shape(1,T0-T1);
  else S=shape(1,T0);
  float width=0.8*hx;
  float height=0.8*hy;
  for(int k=0;k<NCircles_x;k++){
    point2d d0=pmin+point2d(hx*0.5+k*hx,hy*0.5);
    point2d d1=point2d(pmin.x+hx*0.5+k*hx,pmax.y-hy*0.5);
    bool inclusion=false;
    while(inclusion==false){
      //shape S0=kandisky_circle_generator(2.*hx/5.,Nlevels,CircleNpoints)+d0;
      shape S0=kandisky_rectangle_generator(width,height,Nlevels,min_distance)+d0;
      if(S.inclusion_test(S0)==true){
        S.shapeChild.push_back(S0);
        inclusion=true;
      }
    }
    if(NCircles_y>0){
      inclusion=false;
      while(inclusion==false){
        shape S0=kandisky_rectangle_generator(width,height,Nlevels,min_distance)+d1;
        if(S.inclusion_test(S0)==true){
          S.shapeChild.push_back(S0);
          inclusion=true;
        }
      }
    }
  }
  for(int k=0;k<NCircles_y;k++){
    point2d d0=pmin+point2d(hx*0.5,hx+hy*0.5+k*hy);
    point2d d1=point2d(pmax.x-hx*0.5,hx+pmin.y+hy*0.5+k*hy);
    bool inclusion=false;
    while(inclusion==false){
      shape S0=kandisky_rectangle_generator(width,height,Nlevels,min_distance)+d0;
      if(S.inclusion_test(S0)==true){
        S.shapeChild.push_back(S0);
        inclusion=true;
      }
    }
    if(NCircles_x>0){
      inclusion=false;
      while(inclusion==false){
        shape S0=kandisky_rectangle_generator(width,height,Nlevels,min_distance)+d1;
        if(S.inclusion_test(S0)==true){
          S.shapeChild.push_back(S0);
          inclusion=true;
        }
      }
    }
  }

  return S;

}

triangulation leaf::T()
{
  triangulation T;
  vector<point2d> pV;
  vector<float> rV;
  pV.push_back(p0);
  rV.push_back(r0);
  float dan=(a1-a0)/Niter;
  float drn=(r1-r0)/Niter;
  float a=a0;
  float r=r0;
  point2d v_=point2d(sin(a0),cos(a0));
  for(int k=0;k<Niter;k++){
    if(r<0.01) r=0.01;
     p0=p0+v_*dt;
     pV.push_back(p0);
     rV.push_back(r);
     //printf("p0=(%lf,%lf)\n",p0.x,p0.y);
     float w=(float) rand()/RAND_MAX;
     float dr=w*drl+(1-w)*drr+drn;
     r+=dr;
     w=(float) rand()/RAND_MAX;
     float da=w*dal+(1-w)*dar+dan;
     a+=da;
     if(a0==0 && a1==0){
       float co,si=sin(a);
       if((p0.x<0 && si<0.) || (p0.x>0 && si>0.)){
        co=cos(a);
        si=si-p0.x;
        a=atan2(si,co);
       }
     }
     v_=point2d(sin(a),cos(a));
  }
 // triangulation T1(T0,p0,dt,0,-dangle+angle0,dangle+angle0,0.05*scale,-0.003*scale,0.0*scale,Niter);
 return triangulation(T0,pV,rV);

//  // WE ADD A CIRCLE AT THE END
//  int Nf=T1.p.size()-1;
//  //printf("Nf=%d\n",Nf);
//  if (Nf<2) return triangulation();
//  point2d p2=(T1.p[Nf]+T1.p[Nf-1])*0.5;
//  int Np=12; //2+(100.*dt);
//  float radius=(T1.p[Nf]-T1.p[Nf-1]).norm()*0.5;
//  triangulation T2=T1+triangulation(p2,radius,Np);
  //triangulation T2=T1+triangulation(T1.p[Nf-2],0.02, Np )+triangulation(T1.p[Nf-1],0.03, Np )+triangulation(T1.p[Nf],0.04, Np ) ;

//  T2.clean();
//  T2.update_n();
//  T2.rebuildOnlyContour(1e-10);

//  return T2;

}

//leaf(){p0.x=p0.y=0.; dt=0.05; a0=a1=0.; dal=-0.1; dar=0.1; r=0.05; drl=-0.01; drr=0.01; min_r=0.01; Niter=50; T=build_triangulation();}

leaf::leaf(int type){
  switch(type) {
    case 0 :
    {
      p0.x=p0.y=0.; dt=0.05; a0=a1=0.; dal=-0.1; dar=0.1;  drl=-0.0015; drr=0.; min_r=0.01; Niter=25;
      r0=r1=0.05+0.05*rand()/RAND_MAX;
      break;
    }
    case 1 :
    {
      p0.x=p0.y=0.; dt=0.05; a0=a1=0.; dal=-0.1; dar=0.1; r0=r1=0.05; drl=-0.0075; drr=0.; min_r=0.01; Niter=10;
      break;
    }
    case 2 :
    {
      p0.x=p0.y=0.; dt=0.05; a0=a1=0.; dal=-0.1; dar=0.1; r0=r1=0.05; drl=-0.0075; drr=0.; min_r=0.01; Niter=10;
      break;
    }
    default :
    {
      leaf();
      break;
    }
  }
}

leaf::leaf(int type,triangulation &T1,int indexTriangle,int indexSegment){
    //printf("indexTriangle=%d,indexSegment=%d\n",indexTriangle,indexSegment);
    p0.x=p0.y=0.; dt=0.05;  dal=-0.1; dar=0.1; drl=-0.0015; drr=0.0015; min_r=0.005; Niter=10;
    float max_segment_length=T1.max_segment_length(indexTriangle);
    point2d p=T1.p[T1.t[indexTriangle].v[(indexSegment+1)%3]]-T1.p[T1.t[indexTriangle].v[(indexSegment+0)%3]];
    a0=atan2(-p.y,-p.x);
    r1=min_r;
    if(type==0){ a1=a0/2.; r0=max_segment_length/2.; Niter=1+180*r0;}
    else if(type==1){ a1=0.; r0=max_segment_length/2.; Niter=1+180*r0;}
    else if(type==2){ a1=a0/2.; r0=max_segment_length/4.; Niter=1+180*r0;}
    else if(type==3){ a1=0.; r0=max_segment_length/4.; Niter=2+360*r0;}
    else if(type==4){ a1=a0/2.; r0=max_segment_length/2.; Niter=2+360*r0;}
    else if(type==5){ a1=0.; r0=max_segment_length/2.; Niter=2+360*r0;}
    else if(type==6){ a1=a0/2.; r0=max_segment_length/4.; Niter=2+360*r0;}
    else if(type==7){ a1=0.; r0=max_segment_length/4.; Niter=2+360*r0;}

}

/// CONTOURING A SHAPE WITH ANOTHER SHAPE TYPE
void contouring(shape &S0,shape &S,float min_distance,int shape_type,int Ntries,float distance,float zoom){
  vector<int> SelectedTriangles, SelectedSegments;
  vector<int> SelectedPoints;
  if(min_distance>0.) S.T.boundary_segments(min_distance,SelectedTriangles,SelectedSegments,SelectedPoints);
  else{
    if(S.T.t.size()>1){
      SelectedTriangles.push_back(S.T.t.size()-2);
      SelectedSegments.push_back(2);
    }
  }
  for(int m=0;m<(int) SelectedTriangles.size();m++){
    int Nmax=0;
    float scale=exp(-log(2.)/Ntries),zoom2=zoom;
    while(Nmax++<Ntries){
      shape S1=shape_generator().new_shape(shape_type,S.T,SelectedTriangles[m],SelectedSegments[m],distance,zoom2);
      if(S1.T.p.size()>0 ){
        if(S0.intersection_test(S1)==false){
          //printf("1\n");
          S.shapeChild.push_back(S1);
          S0.extrema_update();
          break;
        }
      }
      zoom2*=scale;
    }
  }
}

/// ADD TO A COLLECTION OF SHAPES TRIANGULATIONS STORED IN A REPOSITORY
void ReadShapeDataBase(vector <shape> &SDB,char RepositoryName[]){
  char name[400];
  //sprintf(name,"dir \"%s\" /TW /ON /B > TriangulationList.txt",RepositoryName);
  //printf("%s\n",name);
  //system(name);
  //system("pause");
  FILE *f;
  //f = fopen("TriangulationList.txt", "r");
  sprintf(name,"%sTriangulationList.txt",RepositoryName);
  printf("name=%s\n",name);
  f = fopen(name, "r");
  if (f == NULL){
    printf("\nNo se pueden leer los ficheros de triangulaciones\n");
    //system("pause");
    { int ii_; scanf("%d",&ii_);}
    return ;
  }
  while(feof(f)==0){
    char nameT[200];
    fscanf(f,"%s\n",nameT);
    sprintf(name,"%s%s",RepositoryName,nameT);
    printf("%s   ",name);
    triangulation T(name);
    T.normalize();
    T.fix_triangle_orientation();
    T.update_n();
    T.rebuild();
    T.rebuild();
    T.rebuild();
    T.rebuild();

    shape S(0,T);
    S.extrema_update();
    S.print();
    SDB.push_back(S);
  }

}

/// ADD TO A COLLECTION OF PALETTES GIVEN BY BMP IMAGES STORED IN A REPOSITORY
void ReadPaletteDataBase(vector <imageBMP> &PDB,char RepositoryName[]){
  char name[300];
  //sprintf(name,"dir \"%s\" /TW /ON /B  > PaletteList.txt",RepositoryName);
  //printf("%s\n",name);
  //system(name);
  FILE *f;
  sprintf(name,"%s/PaletteList.txt",RepositoryName);
  printf("PaletteName=%s\n",name);
  //f = fopen("PaletteList.txt", "r");
  f = fopen(name, "r");
  if (f == NULL){
    printf("\nNo se pueden leer los ficheros de Paletas\n");
    //system("pause");
    { int ii_; scanf("%d",&ii_);}
    return ;
  }
  while(feof(f)==0){
    char nameT[200];
    fscanf(f,"%s\n",nameT);
    sprintf(name,"%s%s",RepositoryName,nameT);
    printf("%s\n",name);

    PDB.push_back(imageBMP(name));

//    imageBMP sortHSV=PDB[PDB.size()-1].sortHSV();
//    sortHSV.write(nameT);
  }

}
