/**
 * \file shape.h
 * \brief class  TO manage shape generation
 * \author Luis Alvarez \n \n
*/


#ifndef TREE_GENERATION_H
#define TREE_GENERATION_H
#include "color.h"
#include "shape.h"
#include <iostream>
#include <vector>
#include <math.h>

#define ami_PI 3.14159265358979323846


/** ----------- SHAPE DATABASE ORGANIZATION ----------------------
 shape_database = 0 -> basic shapes : triangle, rectangle, circle, hexagons, ...
 shape_database = 1 -> leaves
 shape_database = 2 -> kandisky circles/squares
 shape_database = 3 -> trees

 if shape_database>999 then the database is read from disk collection using the index "shape_database-1000"

 if shape_database==-1 random shape database choice

*/

class tree_node_generator_parameters{
  public :
  int random_seed;
  vector<short> nodes; /// sequence of nodes in the tree to get the one where the tree_node applies
  int shape_database; /// database of shape to be used
  int shape_index; /// index of shape in the database
  vector<int> shape_selection; /// index of basic shapes used (if empty we use shape_database and shape_index)
  vector<float> shape_selection_scaling; /// uniform scaling to apply to shape_selection vector elements
  int max_number_shapes; /// max number of objects
  float shape_diameter; /// max shape diameter
  int shape_distribution_type; /// way the shapes are distributed
  float angleInit_1; /// initial rotation angle
  float angleFinal_1; /// final rotation angle
  int Nangles_1; /// number of potential rotation angle
  float angleInit_2; /// initial rotation angle
  float angleFinal_2; /// final rotation angle
  int Nangles_2; /// number of potential rotation angle
  int inclusion_father;
  int exclusion_brothers;
  int exclusion_others;
  float min_uniform_scaling; /// min uniform scaling to apply
  float max_uniform_scaling; /// max uniform scaling to apply
  float min_horizontal_scaling; /// min zoom factor in horizontal
  float max_horizontal_scaling; /// min zoom factor in horizontal
  float min_distance_contouring; /// min distance when contouring objects between the original object and the new contouring one
  float max_distance_contouring; /// max distance when contouring objects between the original object and the new contouring one
  float border_thick; /// thick of the triangulation border (if negative no border is included)

  /// RENDER PARAMETERS
  int random_seed_color;
  int palette_index; /// index of the palette used
  float min_transparency;
  float max_transparency;
  float min_light_spot; /** Brightness reduction factor in object light spot effect */
  float max_light_spot;
  int tesselation_palette_index;
  int keepFatherInclusion;

  /// DEFAULT CONSTRUCTOR
  tree_node_generator_parameters();

  /// INCLUDE SHAPES IN ONE NODE OF THE TREE
  void tree_node_generator(shape &S,vector < vector<shape> > &ShapeDataBaseV,vector<imageBMP> &PaletteDataBase);

  /// DEFINE SHAPE DATABASE AND INDEX FROM A SINGLE INDEX NUMBER
  void shape_database_and_index_definition(int m);
  void shape_database_and_index_definition(int m,vector<int> &Database_Index,vector<int> &Shape_Index);

  /// DEFINE A SINGLE INDEX NUMBER FROM SHAPE DATABASE AND INDEX
  int shape_database_and_index_definition_inverse(int shape_database,int shape_index);
  int shape_database_and_index_definition_inverse(int shape_database,int shape_index,vector<int> &Database_Index,vector<int> &Shape_Index);

};

void tree_generator(shape &S,vector < vector<shape> > &ShapeDataBaseV,vector<imageBMP> &PaletteDataBase,
                      vector<tree_node_generator_parameters> &TngpV,float scale);

/// WRITE IN DISK IMAGE PARAMETER CONFIGURATION
void write_image_configuration_parameters(char name[],shape &S,float &width_float,float &height_float,vector<float> &border_thick,
                                          vector<color> &background_color,float &zoom,vector<tree_node_generator_parameters> &TngpV,int &OrderToDrawShapes,float &scale,string &DefaultPalette);
void write_image_configuration_parameters(char name[],shape &S,float &width_float,float &height_float,vector<float> &border_thick,
                                          vector<color> &background_color,float &zoom,vector<tree_node_generator_parameters> &TngpV,int &OrderToDrawShapes,float &scale,
                                          string &DefaultPalette,vector<int> &Database_Index,vector<int> &Shape_Index);

/// READ FROM DISK IMAGE PARAMETER CONFIGURATION
int read_image_configuration_parameters(char name[],shape &S,float &width_float,float &height_float,vector<float> &border_thick,
                                         vector<color> &background_color,float &zoom,vector<tree_node_generator_parameters> &TngpV,int &OrderToDrawShapes,float &scale,string &DefaultPalette);
string read_image_configuration_parameters(char name[],shape &S,float &width_float,float &height_float,vector<float> &border_thick,
                                         vector<color> &background_color,float &zoom,vector<tree_node_generator_parameters> &TngpV,int &OrderToDrawShapes,float &scale,
                                         string &DefaultPalette,vector<int> &Database_Index,vector<int> &Shape_Index );

/// CORRECTION OF BORDER STRIPE SIZE TO ADAPT TO BORDER WITH PERIODICITY
void border_thick_correction(shape &S,float &width_float,float &height_float,vector<float> &border_thick,vector<tree_node_generator_parameters> &TngpV);



shape RandomShapeParameterConfiguration(float &width_float,float &height_float,vector<float> &border_thick,vector<color> &background_color,
                                        vector<tree_node_generator_parameters> &TngpV,
                                        vector<imageBMP> &PaletteDataBase, int palette_index,int &OrderToDrawShapes,float &scale,bool periodicity);
void AddBorder(shape &S,float &width_float,float &height_float,vector<float> &border_thick,vector<color> &background_color,
                                        vector<tree_node_generator_parameters> &TngpV,
                                        vector<imageBMP> &PaletteDataBase, int palette_index);

int read_image_configuration_parametersV0(char name[],shape &S,float &width_float,float &height_float,vector<float> &border_thick,
                                         vector<color> &background_color,float &zoom,vector<tree_node_generator_parameters> &TngpV,int &OrderToDrawShapes,float &scale,string &DefaultPalette);
void build_database_and_shape_index(vector<int> &Database_Index,vector<int> &Shape_Index, int NewShapes);

#endif
